"""Functions for evaluating Zernike polynomials and their derivatives."""

import functools
from math import factorial

import mpmath

from zernipax.backend import (
    cond,
    custom_jvp,
    custom_jvp_with_jit,
    execute_on_cpu,
    fori_loop,
    jit,
    jnp,
    np,
    switch,
)


def jacobi_poly_single(x, n, alpha, beta=0, P_n1=0, P_n2=0):
    """Evaluate Jacobi for single alpha and n pair."""
    c = 2 * n + alpha + beta
    a1 = 2 * n * (c - n) * (c - 2)
    a2 = (c - 1) * (c * (c - 2) * x + (alpha - beta) * (alpha + beta))
    a3 = 2 * (n + alpha - 1) * (n + beta - 1) * c

    a1 = jnp.where(a1 == 0, 1e-6, a1)
    P_n = (a2 * P_n1 - a3 * P_n2) / a1
    # Checks for special cases
    P_n = jnp.where(n < 0, 0, P_n)
    P_n = jnp.where(n == 0, 1, P_n)
    P_n = jnp.where(n == 1, (alpha + 1) + (alpha + beta + 2) * (x - 1) / 2, P_n)
    return P_n


@execute_on_cpu
@functools.partial(jit, static_argnums=3)
def zernike_radial_unique(r, l, m, dr=0):
    """Radial part of zernike polynomials.

    Calculates Radial part of Zernike Polynomials using Jacobi recursion relation
    by getting rid of the redundant calculations for appropriate modes. This version
    is almost the same as zernike_radial_old function but way faster and more
    accurate. First version of this function is zernike_radial_separate which has
    many function for each derivative definition. User can refer that for clarity.


    Parameters
    ----------
    r : ndarray, shape(N,)
        radial coordinates to evaluate basis
    l : ndarray of int, shape(K,)
        radial mode number(s)
    m : ndarray of int, shape(K,)
        azimuthal mode number(s)
    dr : int
        order of derivative (Default = 0)

    Returns
    -------
    out : ndarray, shape(N,K)
        basis function(s) evaluated at specified points

    """
    if dr > 4:
        raise NotImplementedError(
            "Analytic radial derivatives of Zernike polynomials for order>4 "
            + "have not been implemented."
        )

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]
        P_n1 = P_past[1]
        P_n = jnp.zeros((dr + 1, r.size))

        def find_inter_jacobi(dx, args):
            N, alpha, P_n1, P_n2, P_n = args
            P_n = P_n.at[dx, :].set(
                jacobi_poly_single(r_jacobi, N - dx, alpha + dx, dx, P_n1[dx], P_n2[dx])
            )
            return (N, alpha, P_n1, P_n2, P_n)

        # Calculate Jacobi polynomial and derivatives for (m,n)
        _, _, _, _, P_n = fori_loop(
            0, dr + 1, find_inter_jacobi, (N, alpha, P_n1, P_n2, P_n)
        )

        coef = jnp.array(
            [
                1,
                (alpha + N + 1) / 2,
                (alpha + N + 2) * (alpha + N + 1) / 4,
                (alpha + N + 3) * (alpha + N + 2) * (alpha + N + 1) / 8,
                (alpha + N + 4)
                * (alpha + N + 3)
                * (alpha + N + 2)
                * (alpha + N + 1)
                / 16,
            ]
        )
        # TODO: A version without if statements are possible?
        if dr == 0:
            result = (-1) ** N * r**alpha * P_n[0]
        elif dr == 1:
            result = (-1) ** N * (
                alpha * r ** jnp.maximum(alpha - 1, 0) * P_n[0]
                - coef[1] * 4 * r ** (alpha + 1) * P_n[1]
            )
        elif dr == 2:
            result = (-1) ** N * (
                (alpha - 1) * alpha * r ** jnp.maximum(alpha - 2, 0) * P_n[0]
                - coef[1] * 4 * (2 * alpha + 1) * r**alpha * P_n[1]
                + coef[2] * 16 * r ** (alpha + 2) * P_n[2]
            )
        elif dr == 3:
            result = (-1) ** N * (
                (alpha - 2)
                * (alpha - 1)
                * alpha
                * r ** jnp.maximum(alpha - 3, 0)
                * P_n[0]
                - coef[1] * 12 * alpha**2 * r ** jnp.maximum(alpha - 1, 0) * P_n[1]
                + coef[2] * 48 * (alpha + 1) * r ** (alpha + 1) * P_n[2]
                - coef[3] * 64 * r ** (alpha + 3) * P_n[3]
            )
        elif dr == 4:
            result = (-1) ** N * (
                (alpha - 3)
                * (alpha - 2)
                * (alpha - 1)
                * alpha
                * r ** jnp.maximum(alpha - 4, 0)
                * P_n[0]
                - coef[1]
                * 8
                * alpha
                * (2 * alpha**2 - 3 * alpha + 1)
                * r ** jnp.maximum(alpha - 2, 0)
                * P_n[1]
                + coef[2] * 48 * (2 * alpha**2 + 2 * alpha + 1) * r**alpha * P_n[2]
                - coef[3] * 128 * (2 * alpha + 3) * r ** (alpha + 2) * P_n[3]
                + coef[4] * 256 * r ** (alpha + 4) * P_n[4]
            )
        index = jnp.where(
            jnp.logical_and(m == alpha, n == N),
            jnp.arange(1, m.size + 1),
            0,
        )
        idx = jnp.sum(index)
        # needed for proper index
        idx -= 1
        out = out.at[:, idx].set(jnp.where(idx >= 0, result, out.at[:, idx].get()))

        # Shift past values if needed
        mask = N >= 2 + dxs
        P_n2 = jnp.where(mask[:, None], P_n1, P_n2)
        P_n1 = jnp.where(mask[:, None], P_n, P_n1)
        P_past = P_past.at[0, :, :].set(P_n2)
        P_past = P_past.at[1, :, :].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        def find_init_jacobi(dx, args):
            alpha, P_past = args
            P_past = P_past.at[0, dx, :].set(
                jacobi_poly_single(r_jacobi, 0, alpha + dx, beta=dx)
            )
            P_past = P_past.at[1, dx, :].set(
                jacobi_poly_single(r_jacobi, 1, alpha + dx, beta=dx)
            )
            return (alpha, P_past)

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros((2, dr + 1, r.size))
        _, P_past = fori_loop(0, dr + 1, find_init_jacobi, (alpha, P_past))

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    r = jnp.atleast_1d(r)
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)
    dr = int(dr)

    out = jnp.zeros((r.size, m.size))
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)
    dxs = jnp.arange(0, dr + 1)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@execute_on_cpu
@functools.partial(jit, static_argnums=3)
def zernike_radial(r, l, m, dr=0):
    """Radial part of zernike polynomials.

    Calculates Radial part of Zernike Polynomials using Jacobi recursion relation
    by getting rid of the redundant calculations for appropriate modes. This version
    is almost the same as zernike_radial_old function but way faster and more
    accurate. First version of this function is zernike_radial_separate which has
    many function for each derivative definition. User can refer that for clarity.

    There was even faster version of this code but that doesn't have checks
    for duplicate modes

    # Find the index corresponding to the original array
    # I changed arange function to get rid of 0 as index confusion
    # so if index is full of 0s, there is no such mode
    # (FAST BUT NEED A CHECK FOR DUPLICATE MODES)
    index = jnp.where(
        jnp.logical_and(m == alpha, n == N),
        jnp.arange(1, m.size + 1),
        0,
    )
    idx = jnp.sum(index)
    # needed for proper index
    idx -= 1
    result = (-1) ** N * r**alpha * P_n
    out = out.at[:, idx].set(jnp.where(idx >= 0, result, out.at[:, idx].get()))

    Above part replaces the matrix update conducted by following code,

    _, _, _, out = fori_loop(0, m.size, update, (alpha, N, result, out))

    Parameters
    ----------
    r : ndarray, shape(N,)
        radial coordinates to evaluate basis
    l : ndarray of int, shape(K,)
        radial mode number(s)
    m : ndarray of int, shape(K,)
        azimuthal mode number(s)
    dr : int
        order of derivative (Default = 0)

    Returns
    -------
    out : ndarray, shape(N,K)
        basis function(s) evaluated at specified points

    """
    if dr > 4:
        raise NotImplementedError(
            "Analytic radial derivatives of Zernike polynomials for order>4 "
            + "have not been implemented."
        )

    def update(x, args):
        alpha, N, result, out = args
        idx = jnp.where(jnp.logical_and(m[x] == alpha, n[x] == N), x, -1)

        def falseFun(args):
            _, _, out = args
            return out

        def trueFun(args):
            idx, result, out = args
            out = out.at[:, idx].set(result)
            return out

        out = cond(idx >= 0, trueFun, falseFun, (idx, result, out))
        return (alpha, N, result, out)

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]
        P_n1 = P_past[1]
        P_n = jnp.zeros((dr + 1, r.size))

        def find_inter_jacobi(dx, args):
            N, alpha, P_n1, P_n2, P_n = args
            P_n = P_n.at[dx, :].set(
                jacobi_poly_single(r_jacobi, N - dx, alpha + dx, dx, P_n1[dx], P_n2[dx])
            )
            return (N, alpha, P_n1, P_n2, P_n)

        # Calculate Jacobi polynomial and derivatives for (m,n)
        _, _, _, _, P_n = fori_loop(
            0, dr + 1, find_inter_jacobi, (N, alpha, P_n1, P_n2, P_n)
        )

        coef = jnp.array(
            [
                1,
                (alpha + N + 1) / 2,
                (alpha + N + 2) * (alpha + N + 1) / 4,
                (alpha + N + 3) * (alpha + N + 2) * (alpha + N + 1) / 8,
                (alpha + N + 4)
                * (alpha + N + 3)
                * (alpha + N + 2)
                * (alpha + N + 1)
                / 16,
            ]
        )
        # TODO: A version without if statements are possible?
        if dr == 0:
            result = (-1) ** N * r**alpha * P_n[0]
        elif dr == 1:
            result = (-1) ** N * (
                alpha * r ** jnp.maximum(alpha - 1, 0) * P_n[0]
                - coef[1] * 4 * r ** (alpha + 1) * P_n[1]
            )
        elif dr == 2:
            result = (-1) ** N * (
                (alpha - 1) * alpha * r ** jnp.maximum(alpha - 2, 0) * P_n[0]
                - coef[1] * 4 * (2 * alpha + 1) * r**alpha * P_n[1]
                + coef[2] * 16 * r ** (alpha + 2) * P_n[2]
            )
        elif dr == 3:
            result = (-1) ** N * (
                (alpha - 2)
                * (alpha - 1)
                * alpha
                * r ** jnp.maximum(alpha - 3, 0)
                * P_n[0]
                - coef[1] * 12 * alpha**2 * r ** jnp.maximum(alpha - 1, 0) * P_n[1]
                + coef[2] * 48 * (alpha + 1) * r ** (alpha + 1) * P_n[2]
                - coef[3] * 64 * r ** (alpha + 3) * P_n[3]
            )
        elif dr == 4:
            result = (-1) ** N * (
                (alpha - 3)
                * (alpha - 2)
                * (alpha - 1)
                * alpha
                * r ** jnp.maximum(alpha - 4, 0)
                * P_n[0]
                - coef[1]
                * 8
                * alpha
                * (2 * alpha**2 - 3 * alpha + 1)
                * r ** jnp.maximum(alpha - 2, 0)
                * P_n[1]
                + coef[2] * 48 * (2 * alpha**2 + 2 * alpha + 1) * r**alpha * P_n[2]
                - coef[3] * 128 * (2 * alpha + 3) * r ** (alpha + 2) * P_n[3]
                + coef[4] * 256 * r ** (alpha + 4) * P_n[4]
            )
        _, _, _, out = fori_loop(0, m.size, update, (alpha, N, result, out))

        # Shift past values if needed
        mask = N >= 2 + dxs
        P_n2 = jnp.where(mask[:, None], P_n1, P_n2)
        P_n1 = jnp.where(mask[:, None], P_n, P_n1)
        P_past = P_past.at[0, :, :].set(P_n2)
        P_past = P_past.at[1, :, :].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        def find_init_jacobi(dx, args):
            alpha, P_past = args
            P_past = P_past.at[0, dx, :].set(
                jacobi_poly_single(r_jacobi, 0, alpha + dx, beta=dx)
            )
            P_past = P_past.at[1, dx, :].set(
                jacobi_poly_single(r_jacobi, 1, alpha + dx, beta=dx)
            )
            return (alpha, P_past)

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros((2, dr + 1, r.size))
        _, P_past = fori_loop(0, dr + 1, find_init_jacobi, (alpha, P_past))

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    r = jnp.atleast_1d(r)
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)
    dr = int(dr)

    out = jnp.zeros((r.size, m.size))
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)
    dxs = jnp.arange(0, dr + 1)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@functools.partial(jit, static_argnums=3)
def zernike_radial_old_desc(r, l, m, dr=0):
    """Radial part of zernike polynomials.

    Evaluates basis functions using JAX and a stable
    evaluation scheme based on jacobi polynomials and
    binomial coefficients. Generally faster for L>24
    and differentiable, but slower for low resolution.

    Parameters
    ----------
    r : ndarray, shape(N,)
        radial coordinates to evaluate basis
    l : ndarray of int, shape(K,)
        radial mode number(s)
    m : ndarray of int, shape(K,)
        azimuthal mode number(s)
    dr : int
        order of derivative (Default = 0)

    Returns
    -------
    y : ndarray, shape(N,K)
        basis function(s) evaluated at specified points

    """
    m = jnp.abs(m).astype(float)
    alpha = m
    beta = 0
    n = (l - m) // 2
    s = (-1) ** n
    jacobi_arg = 1 - 2 * r**2
    if dr == 0:
        out = r**m * _jacobi(n, alpha, beta, jacobi_arg, 0)
    elif dr == 1:
        f = _jacobi(n, alpha, beta, jacobi_arg, 0)
        df = _jacobi(n, alpha, beta, jacobi_arg, 1)
        out = m * r ** jnp.maximum(m - 1, 0) * f - 4 * r ** (m + 1) * df
    elif dr == 2:
        f = _jacobi(n, alpha, beta, jacobi_arg, 0)
        df = _jacobi(n, alpha, beta, jacobi_arg, 1)
        d2f = _jacobi(n, alpha, beta, jacobi_arg, 2)
        out = (
            (m - 1) * m * r ** jnp.maximum(m - 2, 0) * f
            - 4 * (2 * m + 1) * r**m * df
            + 16 * r ** (m + 2) * d2f
        )
    elif dr == 3:
        f = _jacobi(n, alpha, beta, jacobi_arg, 0)
        df = _jacobi(n, alpha, beta, jacobi_arg, 1)
        d2f = _jacobi(n, alpha, beta, jacobi_arg, 2)
        d3f = _jacobi(n, alpha, beta, jacobi_arg, 3)
        out = (
            (m - 2) * (m - 1) * m * r ** jnp.maximum(m - 3, 0) * f
            - 12 * m**2 * r ** jnp.maximum(m - 1, 0) * df
            + 48 * (m + 1) * r ** (m + 1) * d2f
            - 64 * r ** (m + 3) * d3f
        )
    elif dr == 4:
        f = _jacobi(n, alpha, beta, jacobi_arg, 0)
        df = _jacobi(n, alpha, beta, jacobi_arg, 1)
        d2f = _jacobi(n, alpha, beta, jacobi_arg, 2)
        d3f = _jacobi(n, alpha, beta, jacobi_arg, 3)
        d4f = _jacobi(n, alpha, beta, jacobi_arg, 4)
        out = (
            (m - 3) * (m - 2) * (m - 1) * m * r ** jnp.maximum(m - 4, 0) * f
            - 8 * m * (2 * m**2 - 3 * m + 1) * r ** jnp.maximum(m - 2, 0) * df
            + 48 * (2 * m**2 + 2 * m + 1) * r**m * d2f
            - 128 * (2 * m + 3) * r ** (m + 2) * d3f
            + 256 * r ** (m + 4) * d4f
        )
    else:
        raise NotImplementedError(
            "Analytic radial derivatives of Zernike polynomials for order>4 "
            + "have not been implemented."
        )
    return s * jnp.where((l - m) % 2 == 0, out, 0)


@custom_jvp_with_jit
def zernike_radial_switch(r, l, m, dr=0):
    """Radial part of zernike polynomials.

    Calculates Radial part of Zernike Polynomials using Jacobi recursion relation
    by getting rid of the redundant calculations for appropriate modes.
    https://en.wikipedia.org/wiki/Jacobi_polynomials#Recurrence_relations

    For the derivatives, the following formula is used with above recursion relation,
    https://en.wikipedia.org/wiki/Jacobi_polynomials#Derivatives

    Used formulas are also in the zerike_eval.ipynb notebook in docs.

    This function can be made faster. However, JAX reverse mode AD causes problems.
    In future, we may use vmap() instead of jnp.vectorize() to be able to set dr as
    static argument, and not calculate every derivative even thoguh not asked.

    Parameters
    ----------
    r : ndarray, shape(N,) or scalar
        radial coordinates to evaluate basis
    l : ndarray of int, shape(K,) or integer
        radial mode number(s)
    m : ndarray of int, shape(K,) or integer
        azimuthal mode number(s)
    dr : int
        order of derivative (Default = 0)

    Returns
    -------
    out : ndarray, shape(N,K)
        basis function(s) evaluated at specified points

    """
    dr = jnp.asarray(dr).astype(int)

    branches = [
        _zernike_radial_vectorized,
        _zernike_radial_vectorized_d1,
        _zernike_radial_vectorized_d2,
        _zernike_radial_vectorized_d3,
        _zernike_radial_vectorized_d4,
    ]
    return switch(dr, branches, r, l, m, dr)


@execute_on_cpu
@custom_jvp_with_jit
def zernike_radial_jvp(r, l, m, dr=0):
    """Radial part of zernike polynomials.

    Calculates Radial part of Zernike Polynomials using Jacobi recursion relation
    by getting rid of the redundant calculations for appropriate modes. This version
    is almost the same as zernike_radial_old function but way faster and more
    accurate. First version of this function is zernike_radial_separate which has
    many function for each derivative definition. User can refer that for clarity.

    Parameters
    ----------
    r : ndarray, shape(N,)
        radial coordinates to evaluate basis
    l : ndarray of int, shape(K,)
        radial mode number(s)
    m : ndarray of int, shape(K,)
        azimuthal mode number(s)
    dr : int
        order of derivative (Default = 0)

    Returns
    -------
    out : ndarray, shape(N,K)
        basis function(s) evaluated at specified points

    """
    if dr > 4:
        raise NotImplementedError(
            "Analytic radial derivatives of Zernike polynomials for order>4 "
            + "have not been implemented."
        )

    def update(x, args):
        alpha, N, result, out = args
        idx = jnp.where(jnp.logical_and(m[x] == alpha, n[x] == N), x, -1)

        def falseFun(args):
            _, _, out = args
            return out

        def trueFun(args):
            idx, result, out = args
            out = out.at[:, idx].set(result)
            return out

        out = cond(idx >= 0, trueFun, falseFun, (idx, result, out))
        return (alpha, N, result, out)

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]
        P_n1 = P_past[1]
        P_n = jnp.zeros((dr + 1, r.size))

        def find_inter_jacobi(dx, args):
            N, alpha, P_n1, P_n2, P_n = args
            P_n = P_n.at[dx, :].set(
                jacobi_poly_single(r_jacobi, N - dx, alpha + dx, dx, P_n1[dx], P_n2[dx])
            )
            return (N, alpha, P_n1, P_n2, P_n)

        # Calculate Jacobi polynomial and derivatives for (m,n)
        _, _, _, _, P_n = fori_loop(
            0, dr + 1, find_inter_jacobi, (N, alpha, P_n1, P_n2, P_n)
        )

        coef = jnp.array(
            [
                1,
                (alpha + N + 1) / 2,
                (alpha + N + 2) * (alpha + N + 1) / 4,
                (alpha + N + 3) * (alpha + N + 2) * (alpha + N + 1) / 8,
                (alpha + N + 4)
                * (alpha + N + 3)
                * (alpha + N + 2)
                * (alpha + N + 1)
                / 16,
            ]
        )
        # TODO: A version without if statements are possible?
        if dr == 0:
            result = (-1) ** N * r**alpha * P_n[0]
        elif dr == 1:
            result = (-1) ** N * (
                alpha * r ** jnp.maximum(alpha - 1, 0) * P_n[0]
                - coef[1] * 4 * r ** (alpha + 1) * P_n[1]
            )
        elif dr == 2:
            result = (-1) ** N * (
                (alpha - 1) * alpha * r ** jnp.maximum(alpha - 2, 0) * P_n[0]
                - coef[1] * 4 * (2 * alpha + 1) * r**alpha * P_n[1]
                + coef[2] * 16 * r ** (alpha + 2) * P_n[2]
            )
        elif dr == 3:
            result = (-1) ** N * (
                (alpha - 2)
                * (alpha - 1)
                * alpha
                * r ** jnp.maximum(alpha - 3, 0)
                * P_n[0]
                - coef[1] * 12 * alpha**2 * r ** jnp.maximum(alpha - 1, 0) * P_n[1]
                + coef[2] * 48 * (alpha + 1) * r ** (alpha + 1) * P_n[2]
                - coef[3] * 64 * r ** (alpha + 3) * P_n[3]
            )
        elif dr == 4:
            result = (-1) ** N * (
                (alpha - 3)
                * (alpha - 2)
                * (alpha - 1)
                * alpha
                * r ** jnp.maximum(alpha - 4, 0)
                * P_n[0]
                - coef[1]
                * 8
                * alpha
                * (2 * alpha**2 - 3 * alpha + 1)
                * r ** jnp.maximum(alpha - 2, 0)
                * P_n[1]
                + coef[2] * 48 * (2 * alpha**2 + 2 * alpha + 1) * r**alpha * P_n[2]
                - coef[3] * 128 * (2 * alpha + 3) * r ** (alpha + 2) * P_n[3]
                + coef[4] * 256 * r ** (alpha + 4) * P_n[4]
            )
        _, _, _, out = fori_loop(0, m.size, update, (alpha, N, result, out))

        # Shift past values if needed
        mask = N >= 2 + dxs
        P_n2 = jnp.where(mask[:, None], P_n1, P_n2)
        P_n1 = jnp.where(mask[:, None], P_n, P_n1)
        P_past = P_past.at[0, :, :].set(P_n2)
        P_past = P_past.at[1, :, :].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        def find_init_jacobi(dx, args):
            alpha, P_past = args
            P_past = P_past.at[0, dx, :].set(
                jacobi_poly_single(r_jacobi, 0, alpha + dx, beta=dx)
            )
            P_past = P_past.at[1, dx, :].set(
                jacobi_poly_single(r_jacobi, 1, alpha + dx, beta=dx)
            )
            return (alpha, P_past)

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros((2, dr + 1, r.size))
        _, P_past = fori_loop(0, dr + 1, find_init_jacobi, (alpha, P_past))

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    r = jnp.atleast_1d(r)
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)
    dr = int(dr)

    out = jnp.zeros((r.size, m.size))
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)
    dxs = jnp.arange(0, dr + 1)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@custom_jvp_with_jit
def zernike_radial_switch_gpu(r, l, m, dr=0):
    """Radial part of zernike polynomials.

    Calculates Radial part of Zernike Polynomials using Jacobi recursion relation
    by getting rid of the redundant calculations for appropriate modes.
    https://en.wikipedia.org/wiki/Jacobi_polynomials#Recurrence_relations

    For the derivatives, the following formula is used with above recursion relation,
    https://en.wikipedia.org/wiki/Jacobi_polynomials#Derivatives

    Used formulas are also in the zerike_eval.ipynb notebook in docs.

    This function can be made faster. However, JAX reverse mode AD causes problems.
    In future, we may use vmap() instead of jnp.vectorize() to be able to set dr as
    static argument, and not calculate every derivative even thoguh not asked.

    Parameters
    ----------
    r : ndarray, shape(N,) or scalar
        radial coordinates to evaluate basis
    l : ndarray of int, shape(K,) or integer
        radial mode number(s)
    m : ndarray of int, shape(K,) or integer
        azimuthal mode number(s)
    dr : int
        order of derivative (Default = 0)

    Returns
    -------
    out : ndarray, shape(N,K)
        basis function(s) evaluated at specified points

    """
    dr = jnp.asarray(dr).astype(int)

    branches = [
        _zernike_radial_vectorized_gpu,
        _zernike_radial_vectorized_d1_gpu,
        _zernike_radial_vectorized_d2_gpu,
        _zernike_radial_vectorized_d3_gpu,
        _zernike_radial_vectorized_d4_gpu,
    ]
    return switch(dr, branches, r, l, m, dr)


def find_intermadiate_jacobi(dx, args):
    """Finds Jacobi function and its derivatives for nth loop."""
    r_jacobi, N, alpha, P_n1, P_n2, P_n = args
    P_n = P_n.at[dx].set(
        jacobi_poly_single(r_jacobi, N - dx, alpha + dx, dx, P_n1[dx], P_n2[dx])
    )
    return (r_jacobi, N, alpha, P_n1, P_n2, P_n)


def update_zernike_output(i, args):
    """Updates Zernike radial output, if the mode is in the inputs."""
    m, n, alpha, N, result, out = args
    idx = jnp.where(jnp.logical_and(m[i] == alpha, n[i] == N), i, -1)

    def falseFun(args):
        _, _, out = args
        return out

    def trueFun(args):
        idx, result, out = args
        out = out.at[idx].set(result)
        return out

    out = cond(idx >= 0, trueFun, falseFun, (idx, result, out))
    return (m, n, alpha, N, result, out)


def find_initial_jacobi(dx, args):
    """Finds initial values of Jacobi Polynomial and derivatives."""
    r_jacobi, alpha, P_past = args
    # Jacobi for n=0
    P_past = P_past.at[0, dx].set(jacobi_poly_single(r_jacobi, 0, alpha + dx, beta=dx))
    # Jacobi for n=1
    P_past = P_past.at[1, dx].set(jacobi_poly_single(r_jacobi, 1, alpha + dx, beta=dx))
    return (r_jacobi, alpha, P_past)


@functools.partial(jnp.vectorize, excluded=(1, 2, 3), signature="()->(k)")
def _zernike_radial_vectorized(r, l, m, dr):
    """Calculation of Radial part of Zernike polynomials."""

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]  # Jacobi at N-2
        P_n1 = P_past[1]  # Jacobi at N-1
        P_n = jacobi_poly_single(r_jacobi, N, alpha, 0, P_n1, P_n2)

        # Only calculate the function at dr th index with input r
        result = (-1) ** N * r**alpha * P_n
        # Check if the calculated values is in the given modes
        _, _, _, _, _, out = fori_loop(
            0, m.size, update_zernike_output, (m, n, alpha, N, result, out)
        )

        # Shift past values if needed
        # For derivative order dx, if N is smaller than 2+dx, then only the initial
        # value calculated by find_init_jacobi function will be used. So, if you update
        # P_n's, preceeding values will be wrong.
        mask = N >= 2
        P_n2 = jnp.where(mask, P_n1, P_n2)
        P_n1 = jnp.where(mask, P_n, P_n1)
        # Form updated P_past matrix
        P_past = P_past.at[0].set(P_n2)
        P_past = P_past.at[1].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros(2)
        P_past = P_past.at[0].set(jacobi_poly_single(r_jacobi, 0, alpha, beta=0))
        # Jacobi for n=1
        P_past = P_past.at[1].set(jacobi_poly_single(r_jacobi, 1, alpha, beta=0))

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    # Make inputs 1D arrays in case they aren't
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)

    # From the vectorization, the overall output will be (r.size, m.size)
    out = jnp.zeros(m.size)
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@functools.partial(jnp.vectorize, excluded=(1, 2, 3), signature="()->(k)")
def _zernike_radial_vectorized_d1(r, l, m, dr):
    """First derivative calculation of Radial part of Zernike polynomials."""

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]  # Jacobi at N-2
        P_n1 = P_past[1]  # Jacobi at N-1
        P_n = jnp.zeros(MAXDR + 1)  # Jacobi at N

        # Calculate Jacobi polynomial and derivatives for (alpha,N)
        _, _, _, _, _, P_n = fori_loop(
            0,
            MAXDR + 1,
            find_intermadiate_jacobi,
            (r_jacobi, N, alpha, P_n1, P_n2, P_n),
        )
        # Calculate coefficients for derivatives.
        coef = jnp.array(
            [
                1,
                (alpha + N + 1) / 2,
                (alpha + N + 2) * (alpha + N + 1) / 4,
                (alpha + N + 3) * (alpha + N + 2) * (alpha + N + 1) / 8,
                (alpha + N + 4)
                * (alpha + N + 3)
                * (alpha + N + 2)
                * (alpha + N + 1)
                / 16,
            ]
        )
        # 1th Derivative of Zernike Radial
        result = (-1) ** N * (
            alpha * r ** jnp.maximum(alpha - 1, 0) * P_n[0]
            - coef[1] * 4 * r ** (alpha + 1) * P_n[1]
        )
        # Check if the calculated values is in the given modes
        _, _, _, _, _, out = fori_loop(
            0, m.size, update_zernike_output, (m, n, alpha, N, result, out)
        )

        # Shift past values if needed
        # For derivative order dx, if N is smaller than 2+dx, then only the initial
        # value calculated by find_init_jacobi function will be used. So, if you update
        # P_n's, preceeding values will be wrong.
        mask = N >= 2 + dxs
        P_n2 = jnp.where(mask, P_n1, P_n2)
        P_n1 = jnp.where(mask, P_n, P_n1)
        # Form updated P_past matrix
        P_past = P_past.at[0, :].set(P_n2)
        P_past = P_past.at[1, :].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros((2, MAXDR + 1))
        _, _, P_past = fori_loop(
            0, MAXDR + 1, find_initial_jacobi, (r_jacobi, alpha, P_past)
        )

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    # Make inputs 1D arrays in case they aren't
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)
    dr = jnp.asarray(dr).astype(int)

    # From the vectorization, the overall output will be (r.size, m.size)
    out = jnp.zeros(m.size)
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)

    # This part can be better implemented. Try to make dr as static argument
    # jnp.vectorize doesn't allow it to be static
    MAXDR = 1
    dxs = jnp.arange(0, MAXDR + 1)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@functools.partial(jnp.vectorize, excluded=(1, 2, 3), signature="()->(k)")
def _zernike_radial_vectorized_d2(r, l, m, dr):
    """Second derivative calculation of Radial part of Zernike polynomials."""

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]  # Jacobi at N-2
        P_n1 = P_past[1]  # Jacobi at N-1
        P_n = jnp.zeros(MAXDR + 1)  # Jacobi at N

        # Calculate Jacobi polynomial and derivatives for (alpha,N)
        _, _, _, _, _, P_n = fori_loop(
            0,
            MAXDR + 1,
            find_intermadiate_jacobi,
            (r_jacobi, N, alpha, P_n1, P_n2, P_n),
        )

        # Calculate coefficients for derivatives.
        coef = jnp.array(
            [
                1,
                (alpha + N + 1) / 2,
                (alpha + N + 2) * (alpha + N + 1) / 4,
                (alpha + N + 3) * (alpha + N + 2) * (alpha + N + 1) / 8,
                (alpha + N + 4)
                * (alpha + N + 3)
                * (alpha + N + 2)
                * (alpha + N + 1)
                / 16,
            ]
        )

        result = (-1) ** N * (
            (alpha - 1) * alpha * r ** jnp.maximum(alpha - 2, 0) * P_n[0]
            - coef[1] * 4 * (2 * alpha + 1) * r**alpha * P_n[1]
            + coef[2] * 16 * r ** (alpha + 2) * P_n[2]
        )
        # Check if the calculated values is in the given modes
        _, _, _, _, _, out = fori_loop(
            0, m.size, update_zernike_output, (m, n, alpha, N, result, out)
        )

        # Shift past values if needed
        # For derivative order dx, if N is smaller than 2+dx, then only the initial
        # value calculated by find_init_jacobi function will be used. So, if you update
        # P_n's, preceeding values will be wrong.
        mask = N >= 2 + dxs
        P_n2 = jnp.where(mask, P_n1, P_n2)
        P_n1 = jnp.where(mask, P_n, P_n1)
        # Form updated P_past matrix
        P_past = P_past.at[0, :].set(P_n2)
        P_past = P_past.at[1, :].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros((2, MAXDR + 1))
        _, _, P_past = fori_loop(
            0, MAXDR + 1, find_initial_jacobi, (r_jacobi, alpha, P_past)
        )

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    # Make inputs 1D arrays in case they aren't
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)
    dr = jnp.asarray(dr).astype(int)

    # From the vectorization, the overall output will be (r.size, m.size)
    out = jnp.zeros(m.size)
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)

    # This part can be better implemented. Try to make dr as static argument
    # jnp.vectorize doesn't allow it to be static
    MAXDR = 2
    dxs = jnp.arange(0, MAXDR + 1)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@functools.partial(jnp.vectorize, excluded=(1, 2, 3), signature="()->(k)")
def _zernike_radial_vectorized_d3(r, l, m, dr):
    """Third derivative calculation of Radial part of Zernike polynomials."""

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]  # Jacobi at N-2
        P_n1 = P_past[1]  # Jacobi at N-1
        P_n = jnp.zeros(MAXDR + 1)  # Jacobi at N

        # Calculate Jacobi polynomial and derivatives for (alpha,N)
        _, _, _, _, _, P_n = fori_loop(
            0,
            MAXDR + 1,
            find_intermadiate_jacobi,
            (r_jacobi, N, alpha, P_n1, P_n2, P_n),
        )

        # Calculate coefficients for derivatives.
        coef = jnp.array(
            [
                1,
                (alpha + N + 1) / 2,
                (alpha + N + 2) * (alpha + N + 1) / 4,
                (alpha + N + 3) * (alpha + N + 2) * (alpha + N + 1) / 8,
                (alpha + N + 4)
                * (alpha + N + 3)
                * (alpha + N + 2)
                * (alpha + N + 1)
                / 16,
            ]
        )

        # 3rd Derivative of Zernike Radial
        result = (-1) ** N * (
            (alpha - 2) * (alpha - 1) * alpha * r ** jnp.maximum(alpha - 3, 0) * P_n[0]
            - coef[1] * 12 * alpha**2 * r ** jnp.maximum(alpha - 1, 0) * P_n[1]
            + coef[2] * 48 * (alpha + 1) * r ** (alpha + 1) * P_n[2]
            - coef[3] * 64 * r ** (alpha + 3) * P_n[3]
        )
        # Check if the calculated values is in the given modes
        _, _, _, _, _, out = fori_loop(
            0, m.size, update_zernike_output, (m, n, alpha, N, result, out)
        )

        # Shift past values if needed
        # For derivative order dx, if N is smaller than 2+dx, then only the initial
        # value calculated by find_init_jacobi function will be used. So, if you update
        # P_n's, preceeding values will be wrong.
        mask = N >= 2 + dxs
        P_n2 = jnp.where(mask, P_n1, P_n2)
        P_n1 = jnp.where(mask, P_n, P_n1)
        # Form updated P_past matrix
        P_past = P_past.at[0, :].set(P_n2)
        P_past = P_past.at[1, :].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros((2, MAXDR + 1))
        _, _, P_past = fori_loop(
            0, MAXDR + 1, find_initial_jacobi, (r_jacobi, alpha, P_past)
        )

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    # Make inputs 1D arrays in case they aren't
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)
    dr = jnp.asarray(dr).astype(int)

    # From the vectorization, the overall output will be (r.size, m.size)
    out = jnp.zeros(m.size)
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)

    # This part can be better implemented. Try to make dr as static argument
    # jnp.vectorize doesn't allow it to be static
    MAXDR = 3
    dxs = jnp.arange(0, MAXDR + 1)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@functools.partial(jnp.vectorize, excluded=(1, 2, 3), signature="()->(k)")
def _zernike_radial_vectorized_d4(r, l, m, dr):
    """Fourth derivative calculation of Radial part of Zernike polynomials."""

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]  # Jacobi at N-2
        P_n1 = P_past[1]  # Jacobi at N-1
        P_n = jnp.zeros(MAXDR + 1)  # Jacobi at N

        # Calculate Jacobi polynomial and derivatives for (alpha,N)
        _, _, _, _, _, P_n = fori_loop(
            0,
            MAXDR + 1,
            find_intermadiate_jacobi,
            (r_jacobi, N, alpha, P_n1, P_n2, P_n),
        )

        # Calculate coefficients for derivatives.
        coef = jnp.array(
            [
                1,
                (alpha + N + 1) / 2,
                (alpha + N + 2) * (alpha + N + 1) / 4,
                (alpha + N + 3) * (alpha + N + 2) * (alpha + N + 1) / 8,
                (alpha + N + 4)
                * (alpha + N + 3)
                * (alpha + N + 2)
                * (alpha + N + 1)
                / 16,
            ]
        )

        # 4th Derivative of Zernike Radial
        result = (-1) ** N * (
            (alpha - 3)
            * (alpha - 2)
            * (alpha - 1)
            * alpha
            * r ** jnp.maximum(alpha - 4, 0)
            * P_n[0]
            - coef[1]
            * 8
            * alpha
            * (2 * alpha**2 - 3 * alpha + 1)
            * r ** jnp.maximum(alpha - 2, 0)
            * P_n[1]
            + coef[2] * 48 * (2 * alpha**2 + 2 * alpha + 1) * r**alpha * P_n[2]
            - coef[3] * 128 * (2 * alpha + 3) * r ** (alpha + 2) * P_n[3]
            + coef[4] * 256 * r ** (alpha + 4) * P_n[4]
        )
        # Check if the calculated values is in the given modes
        _, _, _, _, _, out = fori_loop(
            0, m.size, update_zernike_output, (m, n, alpha, N, result, out)
        )

        # Shift past values if needed
        # For derivative order dx, if N is smaller than 2+dx, then only the initial
        # value calculated by find_init_jacobi function will be used. So, if you update
        # P_n's, preceeding values will be wrong.
        mask = N >= 2 + dxs
        P_n2 = jnp.where(mask, P_n1, P_n2)
        P_n1 = jnp.where(mask, P_n, P_n1)
        # Form updated P_past matrix
        P_past = P_past.at[0, :].set(P_n2)
        P_past = P_past.at[1, :].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros((2, MAXDR + 1))
        _, _, P_past = fori_loop(
            0, MAXDR + 1, find_initial_jacobi, (r_jacobi, alpha, P_past)
        )

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    # Make inputs 1D arrays in case they aren't
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)
    dr = jnp.asarray(dr).astype(int)

    # From the vectorization, the overall output will be (r.size, m.size)
    out = jnp.zeros(m.size)
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)

    # This part can be better implemented. Try to make dr as static argument
    # jnp.vectorize doesn't allow it to be static
    MAXDR = 4
    dxs = jnp.arange(0, MAXDR + 1)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@functools.partial(jnp.vectorize, excluded=(1, 2, 3), signature="()->(k)")
def _zernike_radial_vectorized_gpu(r, l, m, dr):
    """Calculation of Radial part of Zernike polynomials."""

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]  # Jacobi at N-2
        P_n1 = P_past[1]  # Jacobi at N-1
        P_n = jacobi_poly_single(r_jacobi, N, alpha, 0, P_n1, P_n2)

        # Check if the calculated values is in the given modes
        mask = jnp.logical_and(m == alpha, n == N)
        result = (-1) ** N * r**alpha * P_n
        out = jnp.where(mask, result, out)

        # Shift past values if needed
        # For derivative order dx, if N is smaller than 2+dx, then only the initial
        # value calculated by find_init_jacobi function will be used. So, if you update
        # P_n's, preceeding values will be wrong.
        mask = N >= 2
        P_n2 = jnp.where(mask, P_n1, P_n2)
        P_n1 = jnp.where(mask, P_n, P_n1)
        # Form updated P_past matrix
        P_past = P_past.at[0].set(P_n2)
        P_past = P_past.at[1].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros(2)
        P_past = P_past.at[0].set(jacobi_poly_single(r_jacobi, 0, alpha, beta=0))
        # Jacobi for n=1
        P_past = P_past.at[1].set(jacobi_poly_single(r_jacobi, 1, alpha, beta=0))

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    # Make inputs 1D arrays in case they aren't
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)

    # From the vectorization, the overall output will be (r.size, m.size)
    out = jnp.zeros(m.size)
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@functools.partial(jnp.vectorize, excluded=(1, 2, 3), signature="()->(k)")
def _zernike_radial_vectorized_d1_gpu(r, l, m, dr):
    """First derivative calculation of Radial part of Zernike polynomials."""

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]  # Jacobi at N-2
        P_n1 = P_past[1]  # Jacobi at N-1
        P_n = jnp.zeros(MAXDR + 1)  # Jacobi at N

        # Calculate Jacobi polynomial and derivatives for (alpha,N)
        _, _, _, _, _, P_n = fori_loop(
            0,
            MAXDR + 1,
            find_intermadiate_jacobi,
            (r_jacobi, N, alpha, P_n1, P_n2, P_n),
        )
        # Calculate coefficients for derivatives.
        coef = jnp.array(
            [
                1,
                (alpha + N + 1) / 2,
                (alpha + N + 2) * (alpha + N + 1) / 4,
                (alpha + N + 3) * (alpha + N + 2) * (alpha + N + 1) / 8,
                (alpha + N + 4)
                * (alpha + N + 3)
                * (alpha + N + 2)
                * (alpha + N + 1)
                / 16,
            ]
        )
        # 1th Derivative of Zernike Radial
        result = (-1) ** N * (
            alpha * r ** jnp.maximum(alpha - 1, 0) * P_n[0]
            - coef[1] * 4 * r ** (alpha + 1) * P_n[1]
        )
        # Check if the calculated values is in the given modes
        mask = jnp.logical_and(m == alpha, n == N)
        out = jnp.where(mask, result, out)

        # Shift past values if needed
        # For derivative order dx, if N is smaller than 2+dx, then only the initial
        # value calculated by find_init_jacobi function will be used. So, if you update
        # P_n's, preceeding values will be wrong.
        mask = N >= 2 + dxs
        P_n2 = jnp.where(mask, P_n1, P_n2)
        P_n1 = jnp.where(mask, P_n, P_n1)
        # Form updated P_past matrix
        P_past = P_past.at[0, :].set(P_n2)
        P_past = P_past.at[1, :].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros((2, MAXDR + 1))
        _, _, P_past = fori_loop(
            0, MAXDR + 1, find_initial_jacobi, (r_jacobi, alpha, P_past)
        )

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    # Make inputs 1D arrays in case they aren't
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)
    dr = jnp.asarray(dr).astype(int)

    # From the vectorization, the overall output will be (r.size, m.size)
    out = jnp.zeros(m.size)
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)

    # This part can be better implemented. Try to make dr as static argument
    # jnp.vectorize doesn't allow it to be static
    MAXDR = 1
    dxs = jnp.arange(0, MAXDR + 1)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@functools.partial(jnp.vectorize, excluded=(1, 2, 3), signature="()->(k)")
def _zernike_radial_vectorized_d2_gpu(r, l, m, dr):
    """Second derivative calculation of Radial part of Zernike polynomials."""

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]  # Jacobi at N-2
        P_n1 = P_past[1]  # Jacobi at N-1
        P_n = jnp.zeros(MAXDR + 1)  # Jacobi at N

        # Calculate Jacobi polynomial and derivatives for (alpha,N)
        _, _, _, _, _, P_n = fori_loop(
            0,
            MAXDR + 1,
            find_intermadiate_jacobi,
            (r_jacobi, N, alpha, P_n1, P_n2, P_n),
        )

        # Calculate coefficients for derivatives.
        coef = jnp.array(
            [
                1,
                (alpha + N + 1) / 2,
                (alpha + N + 2) * (alpha + N + 1) / 4,
                (alpha + N + 3) * (alpha + N + 2) * (alpha + N + 1) / 8,
                (alpha + N + 4)
                * (alpha + N + 3)
                * (alpha + N + 2)
                * (alpha + N + 1)
                / 16,
            ]
        )

        result = (-1) ** N * (
            (alpha - 1) * alpha * r ** jnp.maximum(alpha - 2, 0) * P_n[0]
            - coef[1] * 4 * (2 * alpha + 1) * r**alpha * P_n[1]
            + coef[2] * 16 * r ** (alpha + 2) * P_n[2]
        )
        # Check if the calculated values is in the given modes
        mask = jnp.logical_and(m == alpha, n == N)
        out = jnp.where(mask, result, out)

        # Shift past values if needed
        # For derivative order dx, if N is smaller than 2+dx, then only the initial
        # value calculated by find_init_jacobi function will be used. So, if you update
        # P_n's, preceeding values will be wrong.
        mask = N >= 2 + dxs
        P_n2 = jnp.where(mask, P_n1, P_n2)
        P_n1 = jnp.where(mask, P_n, P_n1)
        # Form updated P_past matrix
        P_past = P_past.at[0, :].set(P_n2)
        P_past = P_past.at[1, :].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros((2, MAXDR + 1))
        _, _, P_past = fori_loop(
            0, MAXDR + 1, find_initial_jacobi, (r_jacobi, alpha, P_past)
        )

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    # Make inputs 1D arrays in case they aren't
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)
    dr = jnp.asarray(dr).astype(int)

    # From the vectorization, the overall output will be (r.size, m.size)
    out = jnp.zeros(m.size)
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)

    # This part can be better implemented. Try to make dr as static argument
    # jnp.vectorize doesn't allow it to be static
    MAXDR = 2
    dxs = jnp.arange(0, MAXDR + 1)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@functools.partial(jnp.vectorize, excluded=(1, 2, 3), signature="()->(k)")
def _zernike_radial_vectorized_d3_gpu(r, l, m, dr):
    """Third derivative calculation of Radial part of Zernike polynomials."""

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]  # Jacobi at N-2
        P_n1 = P_past[1]  # Jacobi at N-1
        P_n = jnp.zeros(MAXDR + 1)  # Jacobi at N

        # Calculate Jacobi polynomial and derivatives for (alpha,N)
        _, _, _, _, _, P_n = fori_loop(
            0,
            MAXDR + 1,
            find_intermadiate_jacobi,
            (r_jacobi, N, alpha, P_n1, P_n2, P_n),
        )

        # Calculate coefficients for derivatives.
        coef = jnp.array(
            [
                1,
                (alpha + N + 1) / 2,
                (alpha + N + 2) * (alpha + N + 1) / 4,
                (alpha + N + 3) * (alpha + N + 2) * (alpha + N + 1) / 8,
                (alpha + N + 4)
                * (alpha + N + 3)
                * (alpha + N + 2)
                * (alpha + N + 1)
                / 16,
            ]
        )

        # 3rd Derivative of Zernike Radial
        result = (-1) ** N * (
            (alpha - 2) * (alpha - 1) * alpha * r ** jnp.maximum(alpha - 3, 0) * P_n[0]
            - coef[1] * 12 * alpha**2 * r ** jnp.maximum(alpha - 1, 0) * P_n[1]
            + coef[2] * 48 * (alpha + 1) * r ** (alpha + 1) * P_n[2]
            - coef[3] * 64 * r ** (alpha + 3) * P_n[3]
        )
        # Check if the calculated values is in the given modes
        mask = jnp.logical_and(m == alpha, n == N)
        out = jnp.where(mask, result, out)

        # Shift past values if needed
        # For derivative order dx, if N is smaller than 2+dx, then only the initial
        # value calculated by find_init_jacobi function will be used. So, if you update
        # P_n's, preceeding values will be wrong.
        mask = N >= 2 + dxs
        P_n2 = jnp.where(mask, P_n1, P_n2)
        P_n1 = jnp.where(mask, P_n, P_n1)
        # Form updated P_past matrix
        P_past = P_past.at[0, :].set(P_n2)
        P_past = P_past.at[1, :].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros((2, MAXDR + 1))
        _, _, P_past = fori_loop(
            0, MAXDR + 1, find_initial_jacobi, (r_jacobi, alpha, P_past)
        )

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    # Make inputs 1D arrays in case they aren't
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)
    dr = jnp.asarray(dr).astype(int)

    # From the vectorization, the overall output will be (r.size, m.size)
    out = jnp.zeros(m.size)
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)

    # This part can be better implemented. Try to make dr as static argument
    # jnp.vectorize doesn't allow it to be static
    MAXDR = 3
    dxs = jnp.arange(0, MAXDR + 1)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@functools.partial(jnp.vectorize, excluded=(1, 2, 3), signature="()->(k)")
def _zernike_radial_vectorized_d4_gpu(r, l, m, dr):
    """Fourth derivative calculation of Radial part of Zernike polynomials."""

    def body_inner(N, args):
        alpha, out, P_past = args
        P_n2 = P_past[0]  # Jacobi at N-2
        P_n1 = P_past[1]  # Jacobi at N-1
        P_n = jnp.zeros(MAXDR + 1)  # Jacobi at N

        # Calculate Jacobi polynomial and derivatives for (alpha,N)
        _, _, _, _, _, P_n = fori_loop(
            0,
            MAXDR + 1,
            find_intermadiate_jacobi,
            (r_jacobi, N, alpha, P_n1, P_n2, P_n),
        )

        # Calculate coefficients for derivatives.
        coef = jnp.array(
            [
                1,
                (alpha + N + 1) / 2,
                (alpha + N + 2) * (alpha + N + 1) / 4,
                (alpha + N + 3) * (alpha + N + 2) * (alpha + N + 1) / 8,
                (alpha + N + 4)
                * (alpha + N + 3)
                * (alpha + N + 2)
                * (alpha + N + 1)
                / 16,
            ]
        )

        # 4th Derivative of Zernike Radial
        result = (-1) ** N * (
            (alpha - 3)
            * (alpha - 2)
            * (alpha - 1)
            * alpha
            * r ** jnp.maximum(alpha - 4, 0)
            * P_n[0]
            - coef[1]
            * 8
            * alpha
            * (2 * alpha**2 - 3 * alpha + 1)
            * r ** jnp.maximum(alpha - 2, 0)
            * P_n[1]
            + coef[2] * 48 * (2 * alpha**2 + 2 * alpha + 1) * r**alpha * P_n[2]
            - coef[3] * 128 * (2 * alpha + 3) * r ** (alpha + 2) * P_n[3]
            + coef[4] * 256 * r ** (alpha + 4) * P_n[4]
        )
        # Check if the calculated values is in the given modes
        mask = jnp.logical_and(m == alpha, n == N)
        out = jnp.where(mask, result, out)

        # Shift past values if needed
        # For derivative order dx, if N is smaller than 2+dx, then only the initial
        # value calculated by find_init_jacobi function will be used. So, if you update
        # P_n's, preceeding values will be wrong.
        mask = N >= 2 + dxs
        P_n2 = jnp.where(mask, P_n1, P_n2)
        P_n1 = jnp.where(mask, P_n, P_n1)
        # Form updated P_past matrix
        P_past = P_past.at[0, :].set(P_n2)
        P_past = P_past.at[1, :].set(P_n1)

        return (alpha, out, P_past)

    def body(alpha, out):
        # find l values with m values equal to alpha
        l_alpha = jnp.where(m == alpha, l, 0)
        # find the maximum among them
        L_max = jnp.max(l_alpha)
        # Maximum possible value for n for loop bound
        N_max = (L_max - alpha) // 2

        # First 2 Jacobi Polynomials (they don't need recursion)
        # P_past stores last 2 Jacobi polynomials (and required derivatives)
        # evaluated at given r points
        P_past = jnp.zeros((2, MAXDR + 1))
        _, _, P_past = fori_loop(
            0, MAXDR + 1, find_initial_jacobi, (r_jacobi, alpha, P_past)
        )

        # Loop over every n value
        _, out, _ = fori_loop(
            0, (N_max + 1).astype(int), body_inner, (alpha, out, P_past)
        )
        return out

    # Make inputs 1D arrays in case they aren't
    m = jnp.atleast_1d(m)
    l = jnp.atleast_1d(l)
    dr = jnp.asarray(dr).astype(int)

    # From the vectorization, the overall output will be (r.size, m.size)
    out = jnp.zeros(m.size)
    r_jacobi = 1 - 2 * r**2
    m = jnp.abs(m)
    n = ((l - m) // 2).astype(int)

    # This part can be better implemented. Try to make dr as static argument
    # jnp.vectorize doesn't allow it to be static
    MAXDR = 4
    dxs = jnp.arange(0, MAXDR + 1)

    M_max = jnp.max(m)
    # Loop over every different m value. There is another nested
    # loop which will execute necessary n values.
    out = fori_loop(0, (M_max + 1).astype(int), body, (out))
    return out


@jit
def fourier(theta, m, NFP=1, dt=0):
    """Fourier series.

    Parameters
    ----------
    theta : ndarray, shape(N,)
        poloidal/toroidal coordinates to evaluate basis
    m : ndarray of int, shape(K,)
        poloidal/toroidal mode number(s)
    NFP : int
        number of field periods (Default = 1)
    dt : int
        order of derivative (Default = 0)

    Returns
    -------
    y : ndarray, shape(N,K)
        basis function(s) evaluated at specified points

    """
    theta, m, NFP, dt = map(jnp.asarray, (theta, m, NFP, dt))
    m_pos = (m >= 0).astype(int)
    m_abs = jnp.abs(m) * NFP
    shift = m_pos * jnp.pi / 2 + dt * jnp.pi / 2
    return m_abs**dt * jnp.sin(m_abs * theta + shift)


@custom_jvp
@jit
@jnp.vectorize
def _jacobi(n, alpha, beta, x, dx=0):
    """Jacobi polynomial evaluation.

    Implementation is only correct for non-negative integer coefficients,
    returns 0 otherwise.

    Parameters
    ----------
    n : int, array_like
        Degree of the polynomial.
    alpha : int, array_like
        Parameter
    beta : int, array_like
        Parameter
    x : float, array_like
        Points at which to evaluate the polynomial

    Returns
    -------
    P : ndarray
        Values of the Jacobi polynomial
    """
    # adapted from scipy:
    # https://github.com/scipy/scipy/blob/701ffcc8a6f04509d115aac5e5681c538b5265a2/
    # scipy/special/orthogonal_eval.pxd#L144

    def _jacobi_body_fun(kk, d_p_a_b_x):
        d, p, alpha, beta, x = d_p_a_b_x
        k = kk + 1.0
        t = 2 * k + alpha + beta
        d = (
            (t * (t + 1) * (t + 2)) * (x - 1) * p + 2 * k * (k + beta) * (t + 2) * d
        ) / (2 * (k + alpha + 1) * (k + alpha + beta + 1) * t)
        p = d + p
        return (d, p, alpha, beta, x)

    n, alpha, beta, x = map(jnp.asarray, (n, alpha, beta, x))

    # coefficient for derivative
    coeffs = jnp.array(
        [
            1,
            (alpha + n + 1) / 2,
            (alpha + n + 2) * (alpha + n + 1) / 4,
            (alpha + n + 3) * (alpha + n + 2) * (alpha + n + 1) / 8,
            (alpha + n + 4) * (alpha + n + 3) * (alpha + n + 2) * (alpha + n + 1) / 16,
        ]
    )
    c = coeffs[dx]
    # taking derivative is same as c*jacobi but for shifted n,a,b
    n -= dx
    alpha += dx
    beta += dx

    d = (alpha + beta + 2) * (x - 1) / (2 * (alpha + 1))
    p = d + 1
    d, p, alpha, beta, x = fori_loop(
        0, jnp.maximum(n - 1, 0).astype(int), _jacobi_body_fun, (d, p, alpha, beta, x)
    )
    out = _binom(n + alpha, n) * p
    # should be complex for n<0, but it gets replaced elsewhere so just return 0 here
    out = jnp.where(n < 0, 0, out)
    # other edge cases
    out = jnp.where(n == 0, 1.0, out)
    out = jnp.where(n == 1, 0.5 * (2 * (alpha + 1) + (alpha + beta + 2) * (x - 1)), out)
    return c * out


@jit
@jnp.vectorize
def _binom(n, k):
    """Binomial coefficient.

    Implementation is only correct for positive integer n,k and n>=k

    Parameters
    ----------
    n : int, array-like
        number of things to choose from
    k : int, array-like
        number of things chosen

    Returns
    -------
    val : int, float, array-like
        number of possible combinations
    """
    # adapted from scipy:
    # https://github.com/scipy/scipy/blob/701ffcc8a6f04509d115aac5e5681c538b5265a2/
    # scipy/special/orthogonal_eval.pxd#L68

    n, k = map(jnp.asarray, (n, k))

    def _binom_body_fun(i, b_n):
        b, n = b_n
        num = n + 1 - i
        den = i
        return (b * num / den, n)

    kx = k.astype(int)
    b, n = fori_loop(1, 1 + kx, _binom_body_fun, (1.0, n))
    return b


@_jacobi.defjvp
def _jacobi_jvp(x, xdot):
    (n, alpha, beta, x, dx) = x
    (ndot, alphadot, betadot, xdot, dxdot) = xdot
    f = _jacobi(n, alpha, beta, x, dx)
    df = _jacobi(n, alpha, beta, x, dx + 1)
    # in theory n, alpha, beta, dx aren't differentiable (they're integers)
    # but marking them as non-diff argnums seems to cause escaped tracer values.
    # probably a more elegant fix, but just setting those derivatives to zero seems
    # to work fine.
    return f, df * xdot + 0 * ndot + 0 * alphadot + 0 * betadot + 0 * dxdot


def polyder_vec(p, m, exact=False):
    """Vectorized version of polyder.

    For differentiating multiple polynomials of the same degree

    Parameters
    ----------
    p : ndarray, shape(N,M)
        polynomial coefficients. Each row is 1 polynomial, in descending powers of x,
        each column is a power of x
    m : int >=0
        order of derivative
    exact : bool
        Whether to use exact integer arithmetic (not compatible with JAX, but may be
        needed for very high degree polynomials)

    Returns
    -------
    der : ndarray, shape(N,M)
        polynomial coefficients for derivative in descending order

    """
    if exact:
        return _polyder_exact(p, m)
    else:
        return _polyder_jax(p, m)


def _polyder_exact(p, m):
    m = np.asarray(m, dtype=int)  # order of derivative
    p = np.atleast_2d(p)
    order = p.shape[1] - 1

    D = np.arange(order, -1, -1)
    num = np.array([factorial(i) for i in D], dtype=object)
    den = np.array([factorial(max(i - m, 0)) for i in D], dtype=object)
    D = (num // den).astype(p.dtype)

    p = np.roll(D * p, m, axis=1)
    idx = np.arange(p.shape[1])
    p = np.where(idx < m, 0, p)
    return p


@jit
def _polyder_jax(p, m):
    p = jnp.atleast_2d(jnp.asarray(p))
    m = jnp.asarray(m).astype(int)
    order = p.shape[1] - 1
    D = jnp.arange(order, -1, -1)

    def body(i, Di):
        return Di * jnp.maximum(D - i, 1)

    D = fori_loop(0, m, body, jnp.ones_like(D))

    p = jnp.roll(D * p, m, axis=1)
    idx = jnp.arange(p.shape[1])
    p = jnp.where(idx < m, 0, p)

    return p


def polyval_vec(p, x, prec=None):
    """Evaluate a polynomial at specific values.

    Vectorized for evaluating multiple polynomials of the same degree.

    Parameters
    ----------
    p : ndarray, shape(N,M)
        Array of coefficient for N polynomials of order M.
        Each row is one polynomial, given in descending powers of x.
    x : ndarray, shape(K,)
        A number, or 1d array of numbers at
        which to evaluate p. If greater than 1d it is flattened.
    prec : int, optional
        precision to use, in number of decimal places. Default is
        double precision (~16 decimals) which should be enough for
        most cases with L <= 24

    Returns
    -------
    y : ndarray, shape(N,K)
        polynomials evaluated at x.
        Each row corresponds to a polynomial, each column to a value of x

    """
    if prec is not None and prec > 18:
        return _polyval_exact(p, x, prec)
    else:
        return _polyval_jax(p, x)


def _polyval_exact(p, x, prec):
    p = np.atleast_2d(p)
    x = np.atleast_1d(x).flatten()
    # TODO: possibly multithread this bit
    mpmath.mp.dps = prec
    y = np.array([np.asarray(mpmath.polyval(list(pi), x)) for pi in p])
    return y.astype(float)


@jit
def _polyval_jax(p, x):
    p = jnp.atleast_2d(jnp.asarray(p))
    x = jnp.atleast_1d(jnp.asarray(x)).flatten()
    npoly = p.shape[0]  # number of polynomials
    order = p.shape[1]  # order of polynomials
    nx = len(x)  # number of coordinates
    y = jnp.zeros((npoly, nx))

    def body(k, y):
        return y * x + jnp.atleast_2d(p[:, k]).T

    y = fori_loop(0, order, body, y)

    return y.astype(float)


def zernike_radial_coeffs(l, m, exact=True):
    """Polynomial coefficients for radial part of zernike basis.

    The for loop ranges from m to l+1 in steps of 2, as opposed to the
    formula in the zernike_eval notebook. This is to make the coeffs array in
    ascending powers of r, which is more natural for polynomial evaluation.
    So, one should substitute s=(l-k)/s in the formula in the notebook to get
    the coding implementation below.

                                 (-1)^((l-k)/2) * ((l+k)/2)!
    R_l^m(r) = sum_{k=m}^l  -------------------------------------
                             ((l-k)/2)! * ((k+m)/2)! * ((k-m)/2)!

    Parameters
    ----------
    l : ndarray of int, shape(K,)
        radial mode number(s)
    m : ndarray of int, shape(K,)
        azimuthal mode number(s)
    exact : bool
        whether to return exact coefficients with `object` dtype
        or return integer or floating point approximation

    Returns
    -------
    coeffs : ndarray
        Polynomial coefficients for Zernike polynomials, in descending powers of r.

    Notes
    -----
    Integer representation is exact up to l~54, so leaving `exact` arg as False
    can speed up evaluation with no loss in accuracy
    """
    from decimal import Decimal, getcontext

    l = np.atleast_1d(l).astype(int)
    m = np.atleast_1d(np.abs(m)).astype(int)
    lm = np.vstack([l, m]).T
    # for modest to large arrays, faster to find unique values and
    # only evaluate those
    lms, idx = np.unique(lm, return_inverse=True, axis=0)

    if exact:
        # Increase the precision of Decimal operations
        getcontext().prec = 100
    else:
        # Use lower precision for not exact calculations
        getcontext().prec = 15
    npoly = len(lms)
    lmax = np.max(lms[:, 0])
    coeffs = np.zeros((npoly, lmax + 1), dtype=object)
    lm_even = ((lms[:, 0] - lms[:, 1]) % 2 == 0)[:, np.newaxis]
    for ii in range(npoly):
        ll = lms[ii, 0]
        mm = lms[ii, 1]
        for s in range(mm, ll + 1, 2):
            coeffs[ii, s] = Decimal(
                int((-1) ** ((ll - s) // 2) * factorial((ll + s) // 2))
            ) / Decimal(
                int(
                    factorial((ll - s) // 2)
                    * factorial((s + mm) // 2)
                    * factorial((s - mm) // 2)
                )
            )
    c = np.fliplr(np.where(lm_even, coeffs, 0))
    if not exact:
        try:
            c = c.astype(int)
        except OverflowError:
            c = c.astype(float)
    c = c[idx]
    return c


def zernike_radial_poly(r, l, m, dr=0, exact="auto"):
    """Radial part of zernike polynomials.

    Evaluates basis functions using numpy to
    exactly compute the polynomial coefficients
    and Horner's method for low resolution,
    or extended precision arithmetic for high resolution.
    Faster for low resolution, but not differentiable.

    Parameters
    ----------
    r : ndarray, shape(N,)
        radial coordinates to evaluate basis
    l : ndarray of int, shape(K,)
        radial mode number(s)
    m : ndarray of int, shape(K,)
        azimuthal mode number(s)
    dr : int
        order of derivative (Default = 0)
    exact : {"auto", True, False}
        Whether to use exact/extended precision arithmetic. Slower but more accurate.
        "auto" will use higher accuracy when needed.

    Returns
    -------
    y : ndarray, shape(N,K)
        basis function(s) evaluated at specified points

    """
    if exact == "auto":
        exact = np.max(l) > 54
    if exact:
        # this should give accuracy of ~1e-10 in the eval'd polynomials
        lmax = np.max(l)
        prec = int(0.4 * lmax + 8.4)
    else:
        prec = None
    coeffs = zernike_radial_coeffs(l, m, exact=exact)
    coeffs = polyder_vec(coeffs, dr, exact=exact)
    return polyval_vec(coeffs, r, prec=prec).T



"""Classes for basis."""

from abc import ABC, abstractmethod

from zernipax.backend import fori_loop, jit, jnp, np, sign
from zernipax.zernike import fourier, zernike_radial


def flatten_list(x, flatten_tuple=False):
    """Flatten a nested list.

    Parameters
    ----------
    x : list
        nested list of lists to flatten
    flatten_tuple : bool
        Whether to also flatten nested tuples.

    Returns
    -------
    x : list
        flattened input

    """
    types = (list,)
    if flatten_tuple:
        types += (tuple,)
    if isinstance(x, types):
        return [a for i in x for a in flatten_list(i, flatten_tuple)]
    else:
        return [x]


@jit
def copy_coeffs(c_old, modes_old, modes_new, c_new=None):
    """Copy coefficients from one resolution to another."""
    modes_old, modes_new = jnp.atleast_1d(modes_old), jnp.atleast_1d(modes_new)

    if modes_old.ndim == 1:
        modes_old = modes_old.reshape((-1, 1))
    if modes_new.ndim == 1:
        modes_new = modes_new.reshape((-1, 1))

    if c_new is None:
        c_new = jnp.zeros((modes_new.shape[0],))
    c_old, c_new = jnp.asarray(c_old), jnp.asarray(c_new)

    def body(i, c_new):
        mask = (modes_old[i, :] == modes_new).all(axis=1)
        c_new = jnp.where(mask, c_old[i], c_new)
        return c_new

    if c_old.size:
        c_new = fori_loop(0, modes_old.shape[0], body, c_new)
    return c_new


class _Basis(ABC):
    """Basis is an abstract base class for spectral basis sets."""

    def __init__(self):
        self._enforce_symmetry()
        self._sort_modes()
        self._create_idx()
        # ensure things that should be ints are ints
        self._L = int(self._L)
        self._M = int(self._M)
        self._N = int(self._N)
        self._NFP = int(self._NFP)
        self._modes = self._modes.astype(int)

    def _set_up(self):
        """Do things after loading or changing resolution."""
        # Also recreates any attributes not in _io_attrs on load from input file.
        # See IOAble class docstring for more info.
        self._enforce_symmetry()
        self._sort_modes()
        self._create_idx()
        # ensure things that should be ints are ints
        self._L = int(self._L)
        self._M = int(self._M)
        self._N = int(self._N)
        self._NFP = int(self._NFP)
        self._modes = self._modes.astype(int)

    def _enforce_symmetry(self):
        """Enforce stellarator symmetry."""
        assert self.sym in [
            "sin",
            "sine",
            "cos",
            "cosine",
            "even",
            "cos(t)",
            False,
            None,
        ], f"Unknown symmetry type {self.sym}"
        if self.sym in ["cos", "cosine"]:  # cos(m*t-n*z) symmetry
            self._modes = self.modes[
                np.asarray(sign(self.modes[:, 1]) == sign(self.modes[:, 2]))
            ]
        elif self.sym in ["sin", "sine"]:  # sin(m*t-n*z) symmetry
            self._modes = self.modes[
                np.asarray(sign(self.modes[:, 1]) != sign(self.modes[:, 2]))
            ]
        elif self.sym == "even":  # even powers of rho
            self._modes = self.modes[np.asarray(self.modes[:, 0] % 2 == 0)]
        elif self.sym == "cos(t)":  # cos(m*t) terms only
            self._modes = self.modes[np.asarray(sign(self.modes[:, 1]) >= 0)]
        elif self.sym is None:
            self._sym = False

    def _sort_modes(self):
        """Sorts modes for use with FFT."""
        sort_idx = np.lexsort((self.modes[:, 1], self.modes[:, 0], self.modes[:, 2]))
        self._modes = self.modes[sort_idx]

    def _create_idx(self):
        """Create index for use with self.get_idx()."""
        self._idx = {}
        for idx, (L, M, N) in enumerate(self.modes):
            if L not in self._idx:
                self._idx[L] = {}
            if M not in self._idx[L]:
                self._idx[L][M] = {}
            self._idx[L][M][N] = idx

    def get_idx(self, L=0, M=0, N=0, error=True):
        """Get the index of the ``'modes'`` array corresponding to given mode numbers.

        Parameters
        ----------
        L : int
            Radial mode number.
        M : int
            Poloidal mode number.
        N : int
            Toroidal mode number.
        error : bool
            whether to raise exception if mode is not in basis, or return empty array

        Returns
        -------
        idx : ndarray of int
            Index of given mode numbers.

        """
        try:
            return self._idx[L][M][N]
        except KeyError as e:
            if error:
                raise ValueError(
                    "mode ({}, {}, {}) is not in basis {}".format(L, M, N, str(self))
                ) from e
            else:
                return np.array([]).astype(int)

    @abstractmethod
    def _get_modes(self):
        """ndarray: Mode numbers for the basis."""

    @abstractmethod
    def evaluate(
        self, nodes, derivatives=np.array([0, 0, 0]), modes=None, unique=False
    ):
        """Evaluate basis functions at specified nodes.

        Parameters
        ----------
        nodes : ndarray of float, size(num_nodes,3)
            node coordinates, in (rho,theta,zeta)
        derivatives : ndarray of int, shape(3,)
            order of derivatives to compute in (rho,theta,zeta)
        modes : ndarray of in, shape(num_modes,3), optional
            basis modes to evaluate (if None, full basis is used)
        unique : bool, optional
            whether to workload by only calculating for unique values of nodes, modes
            can be faster, but doesn't work with jit or autodiff

        Returns
        -------
        y : ndarray, shape(num_nodes,num_modes)
            basis functions evaluated at nodes

        """

    @abstractmethod
    def change_resolution(self):
        """Change resolution of the basis to the given resolutions."""

    @property
    def L(self):
        """int: Maximum radial resolution."""
        return self.__dict__.setdefault("_L", 0)

    @L.setter
    def L(self, L):
        assert int(L) == L, "Basis Resolution must be an integer!"
        self._L = int(L)

    @property
    def M(self):
        """int:  Maximum poloidal resolution."""
        return self.__dict__.setdefault("_M", 0)

    @M.setter
    def M(self, M):
        assert int(M) == M, "Basis Resolution must be an integer!"
        self._M = int(M)

    @property
    def N(self):
        """int: Maximum toroidal resolution."""
        return self.__dict__.setdefault("_N", 0)

    @N.setter
    def N(self, N):
        assert int(N) == N, "Basis Resolution must be an integer!"
        self._N = int(N)

    @property
    def NFP(self):
        """int: Number of field periods."""
        return self.__dict__.setdefault("_NFP", 1)

    @property
    def sym(self):
        """str: {``'cos'``, ``'sin'``, ``False``} Type of symmetry."""
        return self.__dict__.setdefault("_sym", False)

    @property
    def modes(self):
        """ndarray: Mode numbers [l,m,n]."""
        return self.__dict__.setdefault("_modes", np.array([]).reshape((0, 3)))

    @modes.setter
    def modes(self, modes):
        self._modes = modes

    @property
    def num_modes(self):
        """int: Total number of modes in the spectral basis."""
        return self.modes.shape[0]

    @property
    def spectral_indexing(self):
        """str: Type of indexing used for the spectral basis."""
        return self.__dict__.setdefault("_spectral_indexing", "linear")

    def __repr__(self):
        """Get the string form of the object."""
        return (
            type(self).__name__
            + " at "
            + str(hex(id(self)))
            + " (L={}, M={}, N={}, NFP={}, sym={}, spectral_indexing={})".format(
                self.L, self.M, self.N, self.NFP, self.sym, self.spectral_indexing
            )
        )


class ZernikePolynomial(_Basis):
    """2D basis set for analytic functions in a unit disc.

    Parameters
    ----------
    L : int
        Maximum radial resolution. Use L=-1 for default based on M.
    M : int
        Maximum poloidal resolution.
    sym : {``'cos'``, ``'sin'``, ``False``}
        * ``'cos'`` for cos(m*t-n*z) symmetry
        * ``'sin'`` for sin(m*t-n*z) symmetry
        * ``False`` for no symmetry (Default)
    spectral_indexing : {``'ansi'``, ``'fringe'``}
        Indexing method, default value = ``'ansi'``

        For L=0, all methods are equivalent and give a "chevron" shaped
        basis (only the outer edge of the zernike pyramid of width M).
        For L>0, the indexing scheme defines order of the basis functions:

        ``'ansi'``: ANSI indexing fills in the pyramid with triangles of
        decreasing size, ending in a triangle shape. For L == M,
        the traditional ANSI pyramid indexing is recovered. For L>M, adds rows
        to the bottom of the pyramid, increasing L while keeping M constant,
        giving a "house" shape.

        ``'fringe'``: Fringe indexing fills in the pyramid with chevrons of
        decreasing size, ending in a diamond shape for L=2*M where
        the traditional fringe/U of Arizona indexing is recovered.
        For L > 2*M, adds chevrons to the bottom, making a hexagonal diamond.

    """

    def __init__(self, L, M, sym=False, spectral_indexing="ansi"):
        self.L = L
        self.M = M
        self.N = 0
        self._NFP = 1
        self._sym = sym
        self._spectral_indexing = spectral_indexing

        self._modes = self._get_modes(
            L=self.L, M=self.M, spectral_indexing=self.spectral_indexing
        )

        super().__init__()

    def _get_modes(self, L=-1, M=0, spectral_indexing="ansi"):
        """Get mode numbers for Fourier-Zernike basis functions.

        Parameters
        ----------
        L : int
            Maximum radial resolution.
        M : int
            Maximum poloidal resolution.
        spectral_indexing : {``'ansi'``, ``'fringe'``}
            Indexing method, default value = ``'ansi'``

            For L=0, all methods are equivalent and give a "chevron" shaped
            basis (only the outer edge of the zernike pyramid of width M).
            For L>0, the indexing scheme defines order of the basis functions:

            ``'ansi'``: ANSI indexing fills in the pyramid with triangles of
            decreasing size, ending in a triangle shape. For L == M,
            the traditional ANSI pyramid indexing is recovered. For L>M, adds rows
            to the bottom of the pyramid, increasing L while keeping M constant,
            giving a "house" shape.

            ``'fringe'``: Fringe indexing fills in the pyramid with chevrons of
            decreasing size, ending in a diamond shape for L=2*M where
            the traditional fringe/U of Arizona indexing is recovered.
            For L > 2*M, adds chevrons to the bottom, making a hexagonal diamond.

        Returns
        -------
        modes : ndarray of int, shape(num_modes,3)
            Array of mode numbers [l,m,n].
            Each row is one basis function with modes (l,m,n).

        """
        assert spectral_indexing in [
            "ansi",
            "fringe",
        ], "Unknown spectral_indexing: {}".format(spectral_indexing)
        default_L = {"ansi": M, "fringe": 2 * M}
        L = L if L >= 0 else default_L.get(spectral_indexing, M)
        self.L = L

        if spectral_indexing == "ansi":
            pol_posm = [
                [(m + d, m) for m in range(0, M + 1) if m + d < M + 1]
                for d in range(0, L + 1, 2)
            ]
            if L > M:
                pol_posm += [
                    (l, m)
                    for l in range(M + 1, L + 1)
                    for m in range(0, M + 1)
                    if (l - m) % 2 == 0
                ]

        elif spectral_indexing == "fringe":
            pol_posm = [
                [(m + d // 2, m - d // 2) for m in range(0, M + 1) if m - d // 2 >= 0]
                for d in range(0, L + 1, 2)
            ]
            if L > 2 * M:
                pol_posm += [
                    [(l - m, m) for m in range(0, M + 1)]
                    for l in range(2 * M, L + 1, 2)
                ]

        pol = [
            [(l, m), (l, -m)] if m != 0 else [(l, m)] for l, m in flatten_list(pol_posm)
        ]
        pol = np.array(flatten_list(pol))
        num_pol = len(pol)
        tor = np.zeros((num_pol, 1))

        return np.hstack([pol, tor])

    def evaluate(
        self, nodes, derivatives=np.array([0, 0, 0]), modes=None, unique=False
    ):
        """Evaluate basis functions at specified nodes.

        Parameters
        ----------
        nodes : ndarray of float, size(num_nodes,3)
            Node coordinates, in (rho,theta,zeta).
        derivatives : ndarray of int, shape(num_derivatives,3)
            Order of derivatives to compute in (rho,theta,zeta).
        modes : ndarray of int, shape(num_modes,3), optional
            Basis modes to evaluate (if None, full basis is used).
        unique : bool, optional
            Whether to workload by only calculating for unique values of nodes, modes
            can be faster, but doesn't work with jit or autodiff.

        Returns
        -------
        y : ndarray, shape(num_nodes,num_modes)
            Basis functions evaluated at nodes.

        """
        if modes is None:
            modes = self.modes
        if derivatives[2] != 0:
            return jnp.zeros((nodes.shape[0], modes.shape[0]))
        if not len(modes):
            return np.array([]).reshape((len(nodes), 0))

        r, t, z = nodes.T
        l, m, n = modes.T
        lm = modes[:, :2]

        if unique:
            _, ridx, routidx = np.unique(
                r, return_index=True, return_inverse=True, axis=0
            )
            _, tidx, toutidx = np.unique(
                t, return_index=True, return_inverse=True, axis=0
            )
            _, lmidx, lmoutidx = np.unique(
                lm, return_index=True, return_inverse=True, axis=0
            )
            _, midx, moutidx = np.unique(
                m, return_index=True, return_inverse=True, axis=0
            )
            r = r[ridx]
            t = t[tidx]
            lm = lm[lmidx]
            m = m[midx]

        radial = zernike_radial(r, lm[:, 0], lm[:, 1], dr=derivatives[0])
        poloidal = fourier(t[:, np.newaxis], m, 1, derivatives[1])

        if unique:
            radial = radial[routidx][:, lmoutidx]
            poloidal = poloidal[toutidx][:, moutidx]

        return radial * poloidal

    def change_resolution(self, L, M, sym=None):
        """Change resolution of the basis to the given resolutions.

        Parameters
        ----------
        L : int
            Maximum radial resolution.
        M : int
            Maximum poloidal resolution.
        sym : bool
            Whether to enforce stellarator symmetry.

        Returns
        -------
        None

        """
        if L != self.L or M != self.M or sym != self.sym:
            self.L = L
            self.M = M
            self._sym = sym if sym is not None else self.sym
            self._modes = self._get_modes(
                self.L, self.M, spectral_indexing=self.spectral_indexing
            )
            self._set_up()


class FourierZernikeBasis(_Basis):
    """3D basis set for analytic functions in a toroidal volume.

    Zernike polynomials in the radial & poloidal coordinates, and a Fourier
    series in the toroidal coordinate.

    Parameters
    ----------
    L : int
        Maximum radial resolution. Use L=-1 for default based on M.
    M : int
        Maximum poloidal resolution.
    N : int
        Maximum toroidal resolution.
    NFP : int
        Number of field periods.
    sym : {``'cos'``, ``'sin'``, ``False``}
        * ``'cos'`` for cos(m*t-n*z) symmetry
        * ``'sin'`` for sin(m*t-n*z) symmetry
        * ``False`` for no symmetry (Default)
    spectral_indexing : {``'ansi'``, ``'fringe'``}
        Indexing method, default value = ``'ansi'``

        For L=0, all methods are equivalent and give a "chevron" shaped
        basis (only the outer edge of the zernike pyramid of width M).
        For L>0, the indexing scheme defines order of the basis functions:

        ``'ansi'``: ANSI indexing fills in the pyramid with triangles of
        decreasing size, ending in a triangle shape. For L == M,
        the traditional ANSI pyramid indexing is recovered. For L>M, adds rows
        to the bottom of the pyramid, increasing L while keeping M constant,
        giving a "house" shape.

        ``'fringe'``: Fringe indexing fills in the pyramid with chevrons of
        decreasing size, ending in a diamond shape for L=2*M where
        the traditional fringe/U of Arizona indexing is recovered.
        For L > 2*M, adds chevrons to the bottom, making a hexagonal diamond.

    """

    def __init__(self, L, M, N, NFP=1, sym=False, spectral_indexing="ansi"):
        self.L = L
        self.M = M
        self.N = N
        self._NFP = NFP
        self._sym = sym
        self._spectral_indexing = spectral_indexing

        self._modes = self._get_modes(
            L=self.L, M=self.M, N=self.N, spectral_indexing=self.spectral_indexing
        )

        super().__init__()

    def _get_modes(self, L=-1, M=0, N=0, spectral_indexing="ansi"):
        """Get mode numbers for Fourier-Zernike basis functions.

        Parameters
        ----------
        L : int
            Maximum radial resolution.
        M : int
            Maximum poloidal resolution.
        N : int
            Maximum toroidal resolution.
        spectral_indexing : {``'ansi'``, ``'fringe'``}
            Indexing method, default value = ``'ansi'``

            For L=0, all methods are equivalent and give a "chevron" shaped
            basis (only the outer edge of the zernike pyramid of width M).
            For L>0, the indexing scheme defines order of the basis functions:

            ``'ansi'``: ANSI indexing fills in the pyramid with triangles of
            decreasing size, ending in a triangle shape. For L == M,
            the traditional ANSI pyramid indexing is recovered. For L>M, adds rows
            to the bottom of the pyramid, increasing L while keeping M constant,
            giving a "house" shape.

            ``'fringe'``: Fringe indexing fills in the pyramid with chevrons of
            decreasing size, ending in a diamond shape for L=2*M where
            the traditional fringe/U of Arizona indexing is recovered.
            For L > 2*M, adds chevrons to the bottom, making a hexagonal diamond.

        Returns
        -------
        modes : ndarray of int, shape(num_modes,3)
            Array of mode numbers [l,m,n].
            Each row is one basis function with modes (l,m,n).

        """
        assert spectral_indexing in [
            "ansi",
            "fringe",
        ], "Unknown spectral_indexing: {}".format(spectral_indexing)
        default_L = {"ansi": M, "fringe": 2 * M}
        L = L if L >= 0 else default_L.get(spectral_indexing, M)
        self.L = L

        if spectral_indexing == "ansi":
            pol_posm = [
                [(m + d, m) for m in range(0, M + 1) if m + d < M + 1]
                for d in range(0, L + 1, 2)
            ]
            if L > M:
                pol_posm += [
                    (l, m)
                    for l in range(M + 1, L + 1)
                    for m in range(0, M + 1)
                    if (l - m) % 2 == 0
                ]

        elif spectral_indexing == "fringe":
            pol_posm = [
                [(m + d // 2, m - d // 2) for m in range(0, M + 1) if m - d // 2 >= 0]
                for d in range(0, L + 1, 2)
            ]
            if L > 2 * M:
                pol_posm += [
                    [(l - m, m) for m in range(0, M + 1)]
                    for l in range(2 * M, L + 1, 2)
                ]

        pol = [
            [(l, m), (l, -m)] if m != 0 else [(l, m)] for l, m in flatten_list(pol_posm)
        ]
        pol = np.array(flatten_list(pol))
        num_pol = len(pol)

        pol = np.tile(pol, (2 * N + 1, 1))
        tor = np.atleast_2d(
            np.tile(np.arange(-N, N + 1), (num_pol, 1)).flatten(order="f")
        ).T
        return np.unique(np.hstack([pol, tor]), axis=0)

    def evaluate(
        self, nodes, derivatives=np.array([0, 0, 0]), modes=None, unique=False
    ):
        """Evaluate basis functions at specified nodes.

        Parameters
        ----------
        nodes : ndarray of float, size(num_nodes,3)
            Node coordinates, in (rho,theta,zeta).
        derivatives : ndarray of int, shape(num_derivatives,3)
            Order of derivatives to compute in (rho,theta,zeta).
        modes : ndarray of int, shape(num_modes,3), optional
            Basis modes to evaluate (if None, full basis is used).
        unique : bool, optional
            Whether to workload by only calculating for unique values of nodes, modes
            can be faster, but doesn't work with jit or autodiff.

        Returns
        -------
        y : ndarray, shape(num_nodes,num_modes)
            Basis functions evaluated at nodes.

        """
        if modes is None:
            modes = self.modes
        if not len(modes):
            return np.array([]).reshape((len(nodes), 0))

        # TODO: avoid duplicate calculations when mixing derivatives
        r, t, z = nodes.T
        l, m, n = modes.T
        lm = modes[:, :2]

        if unique:
            # TODO: can avoid this here by using grid.unique_idx etc
            # and adding unique_modes attributes to basis
            _, ridx, routidx = np.unique(
                r, return_index=True, return_inverse=True, axis=0
            )
            _, tidx, toutidx = np.unique(
                t, return_index=True, return_inverse=True, axis=0
            )
            _, zidx, zoutidx = np.unique(
                z, return_index=True, return_inverse=True, axis=0
            )
            _, lmidx, lmoutidx = np.unique(
                lm, return_index=True, return_inverse=True, axis=0
            )
            _, midx, moutidx = np.unique(
                m, return_index=True, return_inverse=True, axis=0
            )
            _, nidx, noutidx = np.unique(
                n, return_index=True, return_inverse=True, axis=0
            )
            r = r[ridx]
            t = t[tidx]
            z = z[zidx]
            lm = lm[lmidx]
            m = m[midx]
            n = n[nidx]

        radial = zernike_radial(r, lm[:, 0], lm[:, 1], dr=derivatives[0])
        poloidal = fourier(t[:, np.newaxis], m, dt=derivatives[1])
        toroidal = fourier(z[:, np.newaxis], n, NFP=self.NFP, dt=derivatives[2])
        if unique:
            radial = radial[routidx][:, lmoutidx]
            poloidal = poloidal[toutidx][:, moutidx]
            toroidal = toroidal[zoutidx][:, noutidx]

        return radial * poloidal * toroidal

    def change_resolution(self, L, M, N, NFP=None, sym=None):
        """Change resolution of the basis to the given resolutions.

        Parameters
        ----------
        L : int
            Maximum radial resolution.
        M : int
            Maximum poloidal resolution.
        N : int
            Maximum toroidal resolution.
        NFP : int
            Number of field periods.
        sym : bool
            Whether to enforce stellarator symmetry.

        Returns
        -------
        None

        """
        self._NFP = NFP if NFP is not None else self.NFP
        if L != self.L or M != self.M or N != self.N or sym != self.sym:
            self.L = L
            self.M = M
            self.N = N
            self._sym = sym if sym is not None else self.sym
            self._modes = self._get_modes(
                self.L, self.M, self.N, spectral_indexing=self.spectral_indexing
            )
            self._set_up()



"""Tests for zernipax.zernike module."""

import numpy as np

from zernipax.backend import jax
from zernipax.zernike import (
    fourier,
    jacobi_poly_single,
    zernike_radial,
    zernike_radial_jvp,
    zernike_radial_old_desc,
    zernike_radial_poly,
    zernike_radial_switch,
    zernike_radial_switch_gpu,
    zernike_radial_unique,
)


def test_zernike_radial():  # noqa: C901
    """Test zernike_radial function, comparing to analytic formulas."""
    # https://en.wikipedia.org/wiki/Zernike_polynomials#Radial_polynomials

    def Z3_1(x, dx=0):
        if dx == 0:
            return 3 * x**3 - 2 * x
        if dx == 1:
            return 9 * x**2 - 2
        if dx == 2:
            return 18 * x
        if dx == 3:
            return np.full_like(x, 18)
        if dx >= 4:
            return np.zeros_like(x)

    def Z4_2(x, dx=0):
        if dx == 0:
            return 4 * x**4 - 3 * x**2
        if dx == 1:
            return 16 * x**3 - 6 * x
        if dx == 2:
            return 48 * x**2 - 6
        if dx == 3:
            return 96 * x
        if dx == 4:
            return np.full_like(x, 96)
        if dx >= 5:
            return np.zeros_like(x)

    def Z6_2(x, dx=0):
        if dx == 0:
            return 15 * x**6 - 20 * x**4 + 6 * x**2
        if dx == 1:
            return 90 * x**5 - 80 * x**3 + 12 * x
        if dx == 2:
            return 450 * x**4 - 240 * x**2 + 12
        if dx == 3:
            return 1800 * x**3 - 480 * x
        if dx == 4:
            return 5400 * x**2 - 480
        if dx == 5:
            return 10800 * x
        if dx == 6:
            return np.full_like(x, 10800)
        if dx >= 7:
            return np.zeros_like(x)

    l = np.array([3, 4, 6, 4])
    m = np.array([1, 2, 2, 2])
    r = np.linspace(0, 1, 11)  # rho coordinates
    max_dr = 4
    desired = {
        dr: np.array([Z3_1(r, dr), Z4_2(r, dr), Z6_2(r, dr), Z4_2(r, dr)]).T
        for dr in range(max_dr + 1)
    }
    functions = [
        zernike_radial,
        zernike_radial_switch,
        zernike_radial_switch_gpu,
        zernike_radial_jvp,
    ]
    for fun in functions:
        radial = {dr: fun(r, l, m, dr) for dr in range(max_dr + 1)}
        for dr in range(max_dr + 1):
            np.testing.assert_allclose(
                radial[dr],
                desired[dr],
                err_msg="Failed in " + str(dr) + "th derivative in " + fun.__name__,
            )

    functions = [
        zernike_radial_old_desc,
        zernike_radial_poly,
    ]
    for fun in functions:
        radial = {dr: fun(r[:, np.newaxis], l, m, dr) for dr in range(max_dr + 1)}
        for dr in range(max_dr + 1):
            np.testing.assert_allclose(
                radial[dr],
                desired[dr],
                err_msg="Failed in " + str(dr) + "th derivative in " + fun.__name__,
            )


def test_jacobi_poly_single():
    """Test Jacobi Polynomial evaluation for special cases."""
    # https://en.wikipedia.org/wiki/Jacobi_polynomials#Special_cases

    def exact(r, n, alpha, beta):
        if n == 0:
            return np.ones_like(r)
        elif n == 1:
            return (alpha + 1) + (alpha + beta + 2) * ((r - 1) / 2)
        elif n == 2:
            a0 = (alpha + 1) * (alpha + 2) / 2
            a1 = (alpha + 2) * (alpha + beta + 3)
            a2 = (alpha + beta + 3) * (alpha + beta + 4) / 2
            z = (r - 1) / 2
            return a0 + a1 * z + a2 * z**2
        elif n < 0:
            return np.zeros_like(r)

    r = np.linspace(0, 1, 11)
    # alpha and beta pairs for test
    pairs = np.array([[2, 3], [3, 0], [1, 1], [10, 4]])
    n_values = np.array([-1, -2, 0, 1, 2])

    for pair in pairs:
        alpha = pair[0]
        beta = pair[1]
        P0 = jacobi_poly_single(r, 0, alpha, beta)
        P1 = jacobi_poly_single(r, 1, alpha, beta)
        desired = {n: exact(r, n, alpha, beta) for n in n_values}
        values = {n: jacobi_poly_single(r, n, alpha, beta, P1, P0) for n in n_values}

        for n in n_values:
            np.testing.assert_allclose(values[n], desired[n], err_msg=n)


def test_fourier():
    """Test Fourier series evaluation."""
    m = np.array([-1, 0, 1])
    t = np.linspace(0, 2 * np.pi, 8, endpoint=False)  # theta coordinates

    correct_vals = np.array([np.sin(t), np.ones_like(t), np.cos(t)]).T
    correct_ders = np.array([np.cos(t), np.zeros_like(t), -np.sin(t)]).T

    values = fourier(t[:, np.newaxis], m, dt=0)
    derivs = fourier(t[:, np.newaxis], m, dt=1)

    np.testing.assert_allclose(values, correct_vals, atol=1e-8)
    np.testing.assert_allclose(derivs, correct_ders, atol=1e-8)


def test_ad():
    """Test automatic differentiation."""
    r = np.linspace(0, 1, 11)
    l = np.array([3, 4, 6])
    m = np.array([1, 2, 2])

    # Functions that can be fully differentiated in forward and reverse mode
    full = [
        zernike_radial_switch,
        zernike_radial_switch_gpu,
        zernike_radial_jvp,
    ]
    for fun in full:
        _ = fun(r, l, m, 0)
        _ = jax.jacfwd(fun)(r, l, m, 0)
        _ = jax.jacrev(fun)(r, l, m, 0)

    _ = zernike_radial_old_desc(r[:, np.newaxis], l, m, 0)
    _ = jax.jacfwd(zernike_radial_old_desc)(r[:, np.newaxis], l, m, 0)
    _ = jax.jacrev(zernike_radial_old_desc)(r[:, np.newaxis], l, m, 0)

    # Functions that can be differentiated only in forward mode
    part = [
        zernike_radial,
        zernike_radial_unique,
    ]
    for fun in part:
        _ = fun(r, l, m, 0)
        _ = jax.jacfwd(fun)(r, l, m, 0)



"""Functions for plotting."""

import matplotlib
import matplotlib.pyplot as plt

from zernipax.backend import np
from zernipax.grid import LinearGrid
from zernipax.zernike import fourier, zernike_radial


def _set_tight_layout(fig):
    # compat layer to deal with API changes in mpl 3.6.0
    if int(matplotlib._version.version.split(".")[1]) < 6:
        fig.set_tight_layout(True)
    else:
        fig.set_layout_engine("tight")


def plot_basis(basis, return_data=False, **kwargs):
    """Plot basis functions.

    Parameters
    ----------
    basis : Basis
        basis to plot
    return_data : bool
        if True, return the data plotted as well as fig,ax
    **kwargs : dict, optional
        Specify properties of the figure, axis, and plot appearance e.g.::

            plot_X(figsize=(4,6),cmap="plasma")

        Valid keyword arguments are:

        * ``figsize``: tuple of length 2, the size of the figure (to be passed to
          matplotlib)
        * ``cmap``: str, matplotlib colormap scheme to use, passed to ax.contourf
        * ``title_fontsize``: integer, font size of the title

    Returns
    -------
    fig : matplotlib.figure.Figure
        Figure being plotted to.
    ax : matplotlib.axes.Axes, ndarray of axes, or dict of axes
        Axes used for plotting. A single axis is used for 1d basis functions,
        2d or 3d bases return an ndarray or dict of axes.    return_data : bool
        if True, return the data plotted as well as fig,ax
    plot_data : dict
        dictionary of the data plotted, only returned if ``return_data=True``

    Examples
    --------
    .. code-block:: python

        from zernipax.plotting import plot_basis
        from zernipax.basis import ZernikePolynomial
        basis = basis = ZernikePolynomial(L=5, M=5)
        fig, ax = plot_basis(basis)

    """
    title_fontsize = kwargs.pop("title_fontsize", None)

    if basis.__class__.__name__ in ["ZernikePolynomial", "FourierZernikeBasis"]:
        lmax = abs(basis.modes[:, 0]).max().astype(int)
        mmax = abs(basis.modes[:, 1]).max().astype(int)

        grid = LinearGrid(rho=100, theta=100, endpoint=True)
        r = grid.nodes[grid.unique_rho_idx, 0]
        v = grid.nodes[grid.unique_theta_idx, 1]

        fig = plt.figure(figsize=kwargs.get("figsize", (3 * mmax, 3 * lmax / 2)))

        plot_data = {"amplitude": [], "rho": r, "theta": v}

        ax = {i: {} for i in range(lmax + 1)}
        ratios = np.ones(2 * (mmax + 1) + 1)
        ratios[-1] = kwargs.get("cbar_ratio", 0.25)
        gs = matplotlib.gridspec.GridSpec(
            lmax + 2, 2 * (mmax + 1) + 1, width_ratios=ratios
        )

        modes = basis.modes[basis.modes[:, 2] == 0]
        plot_data["l"] = basis.modes[:, 0]
        plot_data["m"] = basis.modes[:, 1]
        Zs = basis.evaluate(grid.nodes, modes=modes)
        for i, (l, m) in enumerate(
            zip(modes[:, 0].astype(int), modes[:, 1].astype(int))
        ):
            Z = Zs[:, i].reshape((grid.num_rho, grid.num_theta))
            ax[l][m] = plt.subplot(
                gs[l + 1, m + mmax : m + mmax + 2], projection="polar"
            )
            ax[l][m].set_title("$l={}, m={}$".format(l, m))
            ax[l][m].axis("off")
            im = ax[l][m].contourf(
                v,
                r,
                Z,
                levels=np.linspace(-1, 1, 100),
                cmap=kwargs.get("cmap", "coolwarm"),
            )
            plot_data["amplitude"].append(Zs)

        cb_ax = plt.subplot(gs[:, -1])
        plt.subplots_adjust(right=0.8)
        cbar = fig.colorbar(im, cax=cb_ax)
        cbar.set_ticks(np.linspace(-1, 1, 9))
        fig.suptitle(
            "{}, $L={}$, $M={}$, spectral indexing = {}".format(
                basis.__class__.__name__, basis.L, basis.M, basis.spectral_indexing
            ),
            y=0.98,
            fontsize=title_fontsize,
        )
        _set_tight_layout(fig)
        if return_data:
            return fig, ax, plot_data

        return fig, ax


def plot_mode(mode, rho=100, theta=100, **kwargs):
    """Plot basis functions.

    Parameters
    ----------
    mode : (2,) array
        [L, M] mode to plot
    rho : int, optional
        Number of points in the radial direction
    theta : int, optional
        Number of points in the angular direction
    **kwargs : dict, optional
        Specify properties of the figure, axis, and plot appearance e.g.::

            plot_X(figsize=(4,6),cmap="plasma")

        Valid keyword arguments are:

        * ``figsize``: tuple of length 2, the size of the figure (to be passed to
          matplotlib)
        * ``cmap``: str, matplotlib colormap scheme to use, passed to ax.contourf
        * ``title_fontsize``: integer, font size of the title

    Returns
    -------
    fig : matplotlib.figure.Figure
        Figure being plotted to.
    ax : matplotlib.axes.Axes, ndarray of axes, or dict of axes
        Axes used for plotting. A single axis is used for 1d basis functions,
        2d or 3d bases return an ndarray or dict of axes.    return_data : bool
        if True, return the data plotted as well as fig,ax
    plot_data : dict
        dictionary of the data plotted, only returned if ``return_data=True``

    Examples
    --------
    .. code-block:: python

        from zernipax.plotting import plot_mode
        mode = [3, 1]
        fig, ax = plot_mode(mode)

    """
    if len(mode) == 2:
        L = mode[0]
        M = mode[1]

        grid = LinearGrid(rho=rho, theta=theta, endpoint=True)
        r = grid.nodes[grid.unique_rho_idx, 0]
        v = grid.nodes[grid.unique_theta_idx, 1]

        fig = plt.figure(figsize=kwargs.get("figsize", (8, 8)))

        radial = zernike_radial(r, L, M)
        poloidal = fourier(v, M)

        Z = radial * poloidal

        ax = plt.subplot(1, 2, 1, projection="polar")
        ax.set_title("$l={}, m={}$".format(L, M))
        ax.axis("off")
        im = ax.contourf(
            v,
            r,
            Z,
            levels=np.linspace(-1, 1, 200),
            cmap=kwargs.get("cmap", "coolwarm"),
        )
        fig.colorbar(im, ax=ax, shrink=0.3, ticks=np.linspace(-1, 1, 9))

        return fig, ax


def plot_modes(modes, rho=100, theta=100, **kwargs):
    """Plot basis functions.

    Parameters
    ----------
    modes : (2,N) array
        N different [L, M] modes to plot
    rho : int, optional
        Number of points in the radial direction
    theta : int, optional
        Number of points in the angular direction
    **kwargs : dict, optional
        Specify properties of the figure, axis, and plot appearance e.g.::

            plot_X(figsize=(4,6),cmap="plasma")

        Valid keyword arguments are:

        * ``figsize``: tuple of length 2, the size of the figure (to be passed to
          matplotlib)
        * ``cmap``: str, matplotlib colormap scheme to use, passed to ax.contourf
        * ``title_fontsize``: integer, font size of the title

    Returns
    -------
    fig : matplotlib.figure.Figure
        Figure being plotted to.
    ax : matplotlib.axes.Axes, ndarray of axes, or dict of axes
        Axes used for plotting. A single axis is used for 1d basis functions,
        2d or 3d bases return an ndarray or dict of axes.    return_data : bool
        if True, return the data plotted as well as fig,ax
    plot_data : dict
        dictionary of the data plotted, only returned if ``return_data=True``

    Examples
    --------
    .. code-block:: python

        from zernipax.plotting import plot_modes
        modes = np.array([[4, 2], [3,1], [3,3]])
        fig, ax = plot_modes(modes)

    """
    if modes.shape[1] == 2:
        L = modes[:, 0]
        M = modes[:, 1]

        grid = LinearGrid(rho=rho, theta=theta, endpoint=True)
        r = grid.nodes[grid.unique_rho_idx, 0]
        v = grid.nodes[grid.unique_theta_idx, 1]

        fig = plt.figure(figsize=kwargs.get("figsize", (8, 8)))

        Z = np.zeros((r.size, v.size))
        for l, m in zip(L, M):
            radial = zernike_radial(r, l, m)
            poloidal = fourier(v, m)
            Z += radial * poloidal
        Z = Z / L.size
        ax = plt.subplot(1, 2, 1, projection="polar")
        ax.axis("off")
        im = ax.contourf(
            v,
            r,
            Z,
            levels=np.linspace(-1, 1, 200),
            cmap=kwargs.get("cmap", "coolwarm"),
        )
        fig.colorbar(im, ax=ax, shrink=0.4, ticks=np.linspace(-1, 1, 9))

        return fig, ax


def plot_comparison(
    exact, methods, basis, dx=0, type="absolute", names=None, print_error=False
):
    """Plot comparison of exact and approximate methods."""
    assert type in ["absolute", "relative"], "type must be 'absolute' or 'relative'"
    if names is not None:
        assert len(names) == len(methods), "title must have the same length as methods"

    N = len(methods)
    res = basis.L

    cmap = plt.cm.jet  # define the colormap
    # extract all colors from the .jet map
    cmaplist = [cmap(i) for i in range(cmap.N)]

    # create the new map
    cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
        "Custom cmap", cmaplist, cmap.N
    )

    # define the bins and normalize
    bounds = np.logspace(-16, 0, 17)
    norm = matplotlib.colors.BoundaryNorm(bounds, cmap.N)

    fig, ax = plt.subplots(1, N, squeeze=True, figsize=(N * 5, 4))
    for i in range(N):
        description = names[i] if names is not None else f"Method {i+1}:"
        if dx == 0:
            Zmn = "Z_{nm}(x)"
            Zmn_p = "\\tilde{Z}_{nm}(x)"
        else:
            derv = "^" + str(dx) if dx > 1 else ""
            Zmn = "\\frac{d" + derv + "Z_{nm}(x)}{d x" + derv + "}"
            Zmn_p = "\\frac{d" + derv + "\\tilde{Z}_{nm}(x)}{d x" + derv + "}"
        title = description + "$\\max_{x \\in (0,1)} |" + Zmn + "-" + Zmn_p
        if type == "absolute":
            c = np.max(abs(methods[i] - exact), axis=0) / np.mean(abs(exact))
            title = title + "|$"
        else:
            c = np.max(abs(methods[i] - exact), axis=0)
            title = title + "| / |\\bar{Z}_{lm}|$"
        im = ax[i].scatter(
            basis.modes[:, 0],
            basis.modes[:, 1],
            c=c,
            norm=norm,
            cmap=cmap,
        )

        ax[i].grid(True)
        ax[i].set_xticks(np.arange(0, res + 1, 5))
        ax[i].set_yticks(np.arange(0, res + 1, 5))
        ax[i].set_xlabel("$n$", fontsize=12)
        ax[i].set_ylabel("$m$", fontsize=12)
        ax[i].set_title(title, fontsize=14)
        if print_error:
            ax[i].text(
                0,
                45,
                f"Max error: {np.max(c):.2e}",
                fontsize=13,
            )
            ax[i].text(
                0,
                40,
                f"Mean error: {np.mean(c):.2e}",
                fontsize=13,
            )
    # Create a separate axis for the colorbar
    cbar_ax = fig.add_axes([0.92, 0.15, 0.02, 0.7])  # [left, bottom, width, height]
    cbar = fig.colorbar(im, cax=cbar_ax, ticks=bounds)
    cbar.ax.set_yticklabels(["{:.0e}".format(foo) for foo in bounds])



"""Backend functions for zernipax, with options for JAX or regular numpy."""

import functools
import warnings

import numpy as np
from termcolor import colored

from zernipax import config, set_device

if config.get("device") is None:
    set_device("cpu")
try:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        import jax
        import jax.numpy as jnp
        import jaxlib
        from jax import config as jax_config

        jax_config.update("jax_enable_x64", True)
        if config.get("kind") == "gpu" and len(jax.devices("gpu")) == 0:
            warnings.warn(
                "JAX failed to detect GPU, are you sure you "
                + "installed JAX with GPU support?"
            )
            set_device("cpu")
        x = jnp.linspace(0, 5)
        y = jnp.exp(x)
    use_jax = True
    print(
        f"using JAX backend, jax version={jax.__version__}, "
        + f"jaxlib version={jaxlib.__version__}, dtype={y.dtype}"
    )
    del x, y
except ModuleNotFoundError:
    jnp = np
    x = jnp.linspace(0, 5)
    y = jnp.exp(x)
    use_jax = False
    set_device(kind="cpu")
    warnings.warn(colored("Failed to load JAX", "red"))
    print("Using NumPy backend, version={}, dtype={}".format(np.__version__, y.dtype))
print(
    "Using device: {}, with {:.2f} GB available memory".format(
        config.get("device"), config.get("avail_mem")
    )
)


if use_jax:  # noqa C901
    jit = jax.jit
    fori_loop = jax.lax.fori_loop
    cond = jax.lax.cond
    switch = jax.lax.switch
    while_loop = jax.lax.while_loop
    vmap = jax.vmap
    scan = jax.lax.scan
    select = jax.lax.select
    bincount = jnp.bincount
    from jax import custom_jvp
    from jax.scipy.special import gammaln

    def put(arr, inds, vals):
        """Functional interface for array "fancy indexing".

        Provides a way to do arr[inds] = vals in a way that works with JAX.

        Parameters
        ----------
        arr : array-like
            Array to populate
        inds : array-like of int
            Indices to populate
        vals : array-like
            Values to insert

        Returns
        -------
        arr : array-like
            Input array with vals inserted at inds.

        """
        if isinstance(arr, np.ndarray):
            arr[inds] = vals
            return arr
        return jnp.asarray(arr).at[inds].set(vals)

    def sign(x):
        """Sign function, but returns 1 for x==0.

        Parameters
        ----------
        x : array-like
            array of input values

        Returns
        -------
        y : array-like
            1 where x>=0, -1 where x<0

        """
        x = jnp.atleast_1d(x)
        y = jnp.where(x == 0, 1, jnp.sign(x))
        return y

    def custom_jvp_with_jit(func):
        """Decorator for custom_jvp with jit.

        This decorator is specifically with functions that have the same
        structure as the zernike_radial such as r, l, m, dr, where dr is
        the static argument.
        """

        @functools.partial(
            custom_jvp,
            nondiff_argnums=(3,),
        )
        def dummy(r, l, m, dr):
            return func(r, l, m, dr)

        @dummy.defjvp
        def _dummy_jvp(nondiff_dr, x, xdot):
            """Custom derivative rule for the function.

            This is just the same function called with dx+1.
            """
            (r, l, m) = x
            (rdot, ldot, mdot) = xdot
            f = dummy(r, l, m, nondiff_dr)
            df = dummy(r, l, m, nondiff_dr + 1)
            return f, (df.T * rdot).T + 0 * ldot + 0 * mdot

        return jit(dummy, static_argnums=3)

    def execute_on_cpu(func):
        """Decorator to set default device to CPU for a function.

        Parameters
        ----------
        func : callable
            Function to decorate

        Returns
        -------
        wrapper : callable
            Decorated function that will run always on CPU even if
            there are available GPUs.
        """

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with jax.default_device(jax.devices("cpu")[0]):
                return func(*args, **kwargs)

        return wrapper

else:
    jit = lambda func, *args, **kwargs: func
    custom_jvp_with_jit = lambda func, *args, **kwargs: func
    execute_on_cpu = lambda func: func
    from scipy.special import gammaln  # noqa F401

    def put(arr, inds, vals):
        """Functional interface for array "fancy indexing".

        Provides a way to do arr[inds] = vals in a way that works with JAX.

        Parameters
        ----------
        arr : array-like
            Array to populate
        inds : array-like of int
            Indices to populate
        vals : array-like
            Values to insert

        Returns
        -------
        arr : array-like
            Input array with vals inserted at inds.

        """
        arr[inds] = vals
        return arr

    def sign(x):
        """Sign function, but returns 1 for x==0.

        Parameters
        ----------
        x : array-like
            array of input values

        Returns
        -------
        y : array-like
            1 where x>=0, -1 where x<0

        """
        x = np.atleast_1d(x)
        y = np.where(x == 0, 1, np.sign(x))
        return y

    def fori_loop(lower, upper, body_fun, init_val):
        """Loop from lower to upper, applying body_fun to init_val.

        This version is for the numpy backend, for jax backend see jax.lax.fori_loop

        Parameters
        ----------
        lower : int
            an integer representing the loop index lower bound (inclusive)
        upper : int
            an integer representing the loop index upper bound (exclusive)
        body_fun : callable
            function of type ``(int, a) -> a``.
        init_val : array-like or container
            initial loop carry value of type ``a``

        Returns
        -------
        final_val: array-like or container
            Loop value from the final iteration, of type ``a``.

        """
        val = init_val
        for i in np.arange(lower, upper):
            val = body_fun(i, val)
        return val

    def cond(pred, true_fun, false_fun, *operand):
        """Conditionally apply true_fun or false_fun.

        This version is for the numpy backend, for jax backend see jax.lax.cond

        Parameters
        ----------
        pred: bool
            which branch function to apply.
        true_fun: callable
            Function (A -> B), to be applied if pred is True.
        false_fun: callable
            Function (A -> B), to be applied if pred is False.
        operand: any
            input to either branch depending on pred. The type can be a scalar, array,
            or any pytree (nested Python tuple/list/dict) thereof.

        Returns
        -------
        value: any
            value of either true_fun(operand) or false_fun(operand), depending on the
            value of pred. The type can be a scalar, array, or any pytree (nested
            Python tuple/list/dict) thereof.

        """
        if pred:
            return true_fun(*operand)
        else:
            return false_fun(*operand)

    def switch(index, branches, operand):
        """Apply exactly one of branches given by index.

        If index is out of bounds, it is clamped to within bounds.

        Parameters
        ----------
        index: int
            which branch function to apply.
        branches: Sequence[Callable]
            sequence of functions (A -> B) to be applied based on index.
        operand: any
            input to whichever branch is applied.

        Returns
        -------
        value: any
            output of branches[index](operand)

        """
        index = np.clip(index, 0, len(branches) - 1)
        return branches[index](operand)

    def while_loop(cond_fun, body_fun, init_val):
        """Call body_fun repeatedly in a loop while cond_fun is True.

        Parameters
        ----------
        cond_fun: callable
            function of type a -> bool.
        body_fun: callable
            function of type a -> a.
        init_val: any
            value of type a, a type that can be a scalar, array, or any pytree (nested
            Python tuple/list/dict) thereof, representing the initial loop carry value.

        Returns
        -------
        value: any
            The output from the final iteration of body_fun, of type a.

        """
        val = init_val
        while cond_fun(val):
            val = body_fun(val)
        return val

    def vmap(fun, out_axes=0):
        """A numpy implementation of jax.lax.map whose API is a subset of jax.vmap.

        Like Python's builtin map,
        except inputs and outputs are in the form of stacked arrays,
        and the returned object is a vectorized version of the input function.

        Parameters
        ----------
        fun: callable
            Function (A -> B)
        out_axes: int
            An integer indicating where the mapped axis should appear in the output.

        Returns
        -------
        fun_vmap: callable
            Vectorized version of fun.

        """

        def fun_vmap(fun_inputs):
            return np.stack([fun(fun_input) for fun_input in fun_inputs], axis=out_axes)

        return fun_vmap

    def custom_jvp(fun, *args, **kwargs):
        """Dummy function for custom_jvp without JAX."""
        fun.defjvp = lambda *args, **kwargs: None
        fun.defjvps = lambda *args, **kwargs: None
        return fun



"""Setup/build/install script for ZERNIPAX."""

import os

import versioneer
from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

with open("README.md") as f:
    long_description = f.read()

with open(os.path.join(here, "requirements.txt"), encoding="utf-8") as f:
    requirements = f.read().splitlines()

setup(
    name="zernipax",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description=(
        "Includes functions to generate Zernike modes and Calculate Zernike "
        "polynomials and their derivatives using JAX."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PlasmaControl/FastZernike/",
    author="Yigit Gunsur Elmacioglu, Rory Conlin, Egemen Kolemen",
    author_email="PlasmaControl@princeton.edu",
    license="MIT",
    keywords="zernike polynomials, optics, astrophysics, spectral "
    + "simulation, basis, orthogonal polynomials, parallel computing, "
    + "JAX",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    install_requires=requirements,
    python_requires=">=3.9",
)
